import numpy as np

from .mcda_method import MCDA_method
from ..normalizations import linear_normalization


class CRADIS(MCDA_method):
    def __init__(self, normalization_method = linear_normalization):
        """
        Create the CRADIS method object
        """
        self.normalization_method = normalization_method


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.
        
        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.
        
        Returns
        --------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 
        
        Examples
        ----------
        >>> cradis = CRADIS()
        >>> pref = cradis(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        CRADIS._verify_input_data(matrix, weights, types)
        return CRADIS._cradis(matrix, weights, types, self.normalization_method)


    @staticmethod
    def _cradis(matrix, weights, types, normalization_method):

        # Normalize the decision matrix
        norm_matrix = normalization_method(matrix, types)
        # Create the weighted normalized decision matrix
        v = norm_matrix * weights

        # Determine the ideal and anti-ideal solution

        # calculation of the ideal solution is done by finding the largest value in 
        # weighted normalized matrix v(max)
        # calculation of the anti-ideal solution is done by finding the smallest value in
        # weighted normalized matrix v(min)

        # Calculation of deviations from ideal and anti-ideal solutions
        # Calculating the grades of the deviation of individual alternatives from ideal and 
        # anti-ideal solutions
        Sp = np.sum(np.max(v) - v, axis = 1)
        Sm = np.sum(v - np.min(v), axis = 1)

        # Calculation of the utility function for each alternative in relation to the deviations 
        # from the optimal alternatives

        # Sop is the optimal alternative that has the smallest distance from the ideal solution
        # Som is the optimal alternative that has the greatest distance from the anti-ideal solution

        Sop = np.sum(np.max(v) - np.max(v, axis = 0))
        Som = np.sum(np.max(v, axis = 0) - np.min(v))

        Kp = Sop / Sp
        Km = Sm / Som
        Q = (Kp + Km) / 2

        # The best alternative is the one that has the highest value
        return Q


import numpy as np

from .mcda_method import MCDA_method


class VIKOR(MCDA_method):
    def __init__(self, normalization_method = None, v = 0.5):
        """Create the VIKOR method object.

        Parameters
        -----------
            normalization_method : function
                VIKOR does not use normalization by default, thus `normalization_method` is set to None by default.
                However, you can choose method for normalization of decision matrix chosen `normalization_method` from `normalizations`.
                It is used in a way `normalization_method(X, types)` where `X` is a decision matrix
                and `types` is a vector with criteria types where 1 means profit and -1 means cost.

            v : float
                parameter that is weight of strategy of the majority of criteria (the maximum group utility)
        """
        self.v = v
        self.normalization_method = normalization_method


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        --------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the lowest preference value.

        Examples
        ---------
        >>> vikor = VIKOR(normalization_method = minmax_normalization)
        >>> pref = vikor(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = False) 
        """
        VIKOR._verify_input_data(matrix, weights, types)
        return VIKOR._vikor(matrix, weights, types, self.normalization_method, self.v)


    @staticmethod
    def _vikor(matrix, weights, types, normalization_method, v):
        # Without special normalization method
        if normalization_method == None:

            # Determine the best `fstar` and the worst `fmin` values of all criterion function
            maximums_matrix = np.amax(matrix, axis = 0)
            minimums_matrix = np.amin(matrix, axis = 0)

            fstar = np.zeros(matrix.shape[1])
            fmin = np.zeros(matrix.shape[1])

            # for profit criteria (`types` == 1) and for cost criteria (`types` == -1)
            fstar[types == 1] = maximums_matrix[types == 1]
            fstar[types == -1] = minimums_matrix[types == -1]
            fmin[types == 1] = minimums_matrix[types == 1]
            fmin[types == -1] = maximums_matrix[types == -1]

            # Calculate the weighted matrix
            weighted_matrix = weights * ((fstar - matrix) / (fstar - fmin))
        else:
            # With special normalization method
            norm_matrix = normalization_method(matrix, types)
            fstar = np.amax(norm_matrix, axis = 0)
            fmin = np.amin(norm_matrix, axis = 0)

            # Calculate the weighted matrix
            weighted_matrix = weights * ((fstar - norm_matrix) / (fstar - fmin))

        # Calculate the `S` and `R` values
        S = np.sum(weighted_matrix, axis = 1)
        R = np.amax(weighted_matrix, axis = 1)
        # Calculate the Q values
        Sstar = np.min(S)
        Smin = np.max(S)
        Rstar = np.min(R)
        Rmin = np.max(R)
        Q = v * (S - Sstar) / (Smin - Sstar) + (1 - v) * (R - Rstar) / (Rmin - Rstar)
        return Q


import numpy as np
import pandas as pd

from .additions import rank_preferences


class Sensitivity_analysis_weights_values():
    def __init__(self):
        """
        Create object of `Sensitivity_analysis_weights_values` method.
        """
        pass


    def __call__(self, matrix, weight_values, types, method, j):
        """
        Method for sensitivity analysis. This method determines rankings of alternatives using chosen
        MCDA method name `mcda_name` for the value of criterion `j` weight set as chosen `weight_value`.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with performance values of alternatives. This matrix includes
                data on alternatives in rows considering criteria in columns.

            weight_values : ndarray
                Vector with values to be set as the weight of chosen criterion in the sensitivity analysis procedure in range from 0 to 1.

            types : ndarray
                Vector with criteria types. Types must be equal to 1 for profit criteria and -1
                for cost criteria.

            method : class
                Initialized object of class of chosen MCDA method

            j : int
                Index of the column in decision matrix `matrix` that indicates for which criterion
                the weight is set with chosen value.

        Returns
        --------
            data_sens : DataFrame
                dataframe with rankings calculated for subsequent changes of criterion `j` weight.
                Particular rankings for different weight values of criterion `j` are included in
                subsequent columns of the dataframe.

        Examples
        ----------
        >>> sensitivity_analysis = Sensitivity_analysis_weights_values()
        >>> df_sens = sensitivity_analysis(matrix, weight_values, types, method, j)
        """
        list_alt_names = [r'$A_{' + str(i) + '}$' for i in range(1, matrix.shape[0] + 1)]
        return Sensitivity_analysis_weights_values._sensitivity_analysis_weights_values(self, matrix, weight_values, types, method, list_alt_names, j)


    def _change_weights(self, matrix, weight_value, j):
        """
        Method for criteria weights changing in sensitivity analysis procedure.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with performance values of alternatives. This matrix includes
                data on alternatives in rows considering criteria in columns

            weight_value : float
                Value in range from 0 to 1 to be set as the weight of chosen criterion with index `j`.

            j : int
                Index of the column in decision matrix `matrix` indicating for which criterion
                the weight value is changed.

        Returns
        --------
            new_weights : ndarray
                Vector with criteria weights after changing their values for sensitivity analysis
        """
        
        new_weights = np.ones(matrix.shape[1]) * ((1 - weight_value) / (matrix.shape[1] - 1))
        new_weights[j] = weight_value
        return new_weights


    @staticmethod
    def _sensitivity_analysis_weights_values(self, matrix, weight_values, types, method, list_alt_names, j):
        # Create a dataframe for sensitivity analysis results (rankings)
        data_sens = pd.DataFrame()
        # Assisgn indexes (alternatives symbols) to dataframe `data_sens`
        data_sens['Ai'] = list_alt_names
        
        # Iterate by values of weight change in vector `weight_values`
        for val in weight_values:
            # Change weights using method named `_change_weights` from class `Sensitivity_analysis_weights_values`
            new_weights = self._change_weights(matrix, val, j)

            # Calculate alternatives ranking using selected MCDA method, `matrix`, vector of new weights
            # `new_weights` and criteria types `types`
            if method.__class__.__name__ == 'TOPSIS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'CODAS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'VIKOR':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = False)

            elif method.__class__.__name__ == 'SPOTIS':
                bounds_min = np.amin(matrix, axis = 0)
                bounds_max = np.amax(matrix, axis = 0)
                bounds = np.vstack((bounds_min, bounds_max))
                pref = method(matrix, new_weights, types, bounds)
                rank = rank_preferences(pref, reverse = False)

            elif method.__class__.__name__ == 'EDAS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'MABAC':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'MULTIMOORA':
                rank = method(matrix, new_weights, types)

            elif method.__class__.__name__ == 'WASPAS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'ARAS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'COPRAS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'CRADIS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'MARCOS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'CRADIS':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'SAW':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'PROMETHEE_II':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            elif method.__class__.__name__ == 'PROSA_C':
                pref = method(matrix, new_weights, types)
                rank = rank_preferences(pref, reverse = True)

            else:
                print('There is no available sensitivity analysis for this method')
                return
            
            # Assign calculated ranking to column with value of j criterion weight change in `data_sens` dataframe
            data_sens['{:.2f}'.format(val)] = rank

        # Set index with alternatives symbols in dataframe with sensitivity analysis results (rankings)
        data_sens = data_sens.set_index('Ai')
        return data_sens


import itertools
import numpy as np

from .promethee import PROMETHEE_II



# PROSA Examining Sustainability at the Criteria Level (PROSA-C)
class PROSA_C(PROMETHEE_II):
    def __init__(self):
        """
        Create the PROSA method object
        """
        pass


    def __call__(self, matrix, weights, types, preference_functions = None, p = None, q = None, s = None):
        PROSA_C._verify_input_data(matrix, weights, types)
        
        if preference_functions is None:
            preference_functions = [self._usual_function for pf in range(len(weights))]

        if p is None:
            u = np.sqrt(np.sum(np.square(np.mean(matrix, axis = 0) - matrix), axis = 0) / matrix.shape[0])
            p = 2 * u
            
        if q is None:
            u = np.sqrt(np.sum(np.square(np.mean(matrix, axis = 0) - matrix), axis = 0) / matrix.shape[0])
            q = 0.5 * u

        if s is None:
            s = np.repeat(0.3, len(weights))

        PROSA_C._verify_input_data(matrix, weights, types)
        # Verification of the correctness of given arguments
        if len(preference_functions) != np.shape(matrix)[1]:
            raise ValueError('The list of preference functions must be equal in length to the number of criteria')
        if len(p) != np.shape(matrix)[1]:
            raise ValueError('The length of the vector p must be equal to the number of criteria')
        if len(q) != np.shape(matrix)[1]:
            raise ValueError('The length of the vector q must be equal to the number of criteria')
        if len(s) != np.shape(matrix)[1]:
            raise ValueError('The length of the vector s must be equal to the number of criteria')

        return PROSA_C._prosa_c(self, matrix, weights, types, preference_functions, p, q, s)


    @staticmethod
    def _prosa_c(self, matrix, weights, types, preference_functions, p, q, s):
        """
        Score alternatives provided in the decision matrix `matrix` using criteria `weights` and criteria `types`.
        
        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. The sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.
            preference_functions : list
                List with methods containing preference functions for calculating the
                preference degree for each criterion.
            p : ndarray
                Vector with values representing the threshold of absolute preference.
            q : ndarray
                Vector with values representing the threshold of indifference.
            s : ndarray
                Vector with values of the coefficient sj for the criteria
        
        Returns
        --------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 
        
        Examples
        ----------
        >>> prosa_c = PROSA_C()
        >>> preference_functions = [prosa_c._linear_function for pf in range(len(weights))]
        >>> u = np.sqrt(np.sum(np.square(np.mean(matrix, axis = 0) - matrix), axis = 0) / matrix.shape[0])
        >>> p = 2 * u
        >>> q = 0.5 * u
        >>> s = np.repeat(0.3, len(weights))
        >>> pref = promethee_II(matrix, weights, types, preference_functions, p = p, q = q, s = s)
        >>> rank = rank_preferences(pref, reverse = True)
        """

        m, n = matrix.shape
        phi = np.zeros((m, n))

        # Determination of deviations based on pair-wise comparisons,
        # Application of the preference function,
        # Calculation of a single criterion net outranking flow
        # preference functions are inherited from the PROMETHEE II class
        for j, i, k in itertools.product(range(n), range(m), range(m)):
            phi[i, j] += preference_functions[j](types[j] * (matrix[i, j] - matrix[k, j]), p[j], q[j]) -\
                preference_functions[j](types[j] * (matrix[k, j] - matrix[i, j]), p[j], q[j])

        
        phi = phi / (m - 1)
        # Calculation of a global (overall) net outranking flow,
        # phi_net is the weighted sum of net flow for each criterion
        phi_net = np.sum(phi * weights, axis = 1)

        # Calculate the value of a mean absolute deviation in a weighted form,
        # where the sustainability (compensation) coefficient was taken into consideration,
        # where s denotes the sustainability (compensation) coefficient for a criterion j.
        WMAD = np.sum(np.abs(phi_net.reshape(-1, 1) - phi) * weights * s, axis = 1)

        # The final evaluation of alternatives (PSV_net) (PROSA net Sustainable Value), is calculated
        PSV_net = phi_net - WMAD

        return PSV_net


import unittest
import numpy as np
from scipy.stats import pearsonr
from pyrepo_mcda import correlations as corrs


# Test for Spearman rank correlation coefficient
class Test_Spearman(unittest.TestCase):

    def test_spearman(self):
        """Test based on paper Sałabun, W., & Urbaniak, K. (2020, June). A new coefficient of rankings similarity 
        in decision-making problems. In International Conference on Computational Science 
        (pp. 632-645). Springer, Cham."""

        R = np.array([1, 2, 3, 4, 5])
        Q = np.array([1, 3, 2, 4, 5])
        test_result = corrs.spearman(R, Q)
        real_result = 0.9
        self.assertEqual(test_result, real_result)


# Test for Weighted Spearman rank correlation coefficient
class Test_Weighted_Spearman(unittest.TestCase):

    def test_weighted_spearman(self):
        """Test based on paper Sałabun, W., & Urbaniak, K. (2020, June). A new coefficient of rankings similarity 
        in decision-making problems. In International Conference on Computational Science 
        (pp. 632-645). Springer, Cham."""

        R = np.array([1, 2, 3, 4, 5])
        Q = np.array([1, 3, 2, 4, 5])
        test_result = corrs.weighted_spearman(R, Q)
        real_result = 0.8833
        self.assertEqual(np.round(test_result, 4), real_result)


# Test for Similarity rank coefficient WS
class Test_WS(unittest.TestCase):

    def test_ws(self):
        """Test based on paper Sałabun, W., & Urbaniak, K. (2020, June). A new coefficient of rankings similarity 
        in decision-making problems. In International Conference on Computational Science 
        (pp. 632-645). Springer, Cham."""

        R = np.array([1, 2, 3, 4, 5])
        Q = np.array([1, 3, 2, 4, 5])
        test_result = corrs.WS_coeff(R, Q)
        real_result = 0.8542
        self.assertEqual(np.round(test_result, 4), real_result)


# Test for Pearson correlation coefficient
class Test_Pearson(unittest.TestCase):

    def test_pearson(self):
        """Test based on paper Sałabun, W., & Urbaniak, K. (2020, June). A new coefficient of rankings similarity 
        in decision-making problems. In International Conference on Computational Science 
        (pp. 632-645). Springer, Cham."""

        R = np.array([1, 2, 3, 4, 5])
        Q = np.array([1, 3, 2, 4, 5])
        test_result = corrs.pearson_coeff(R, Q)
        real_result, _ = pearsonr(R, Q)
        self.assertEqual(test_result, real_result)


def main():
    test_spearman_coeff = Test_Spearman()
    test_spearman_coeff.test_spearman()

    test_weighted_spearman_coeff = Test_Weighted_Spearman()
    test_weighted_spearman_coeff.test_weighted_spearman()

    test_pearson_coeff = Test_Pearson()
    test_pearson_coeff.test_pearson()

    test_ws = Test_WS()
    test_ws.test_ws()


if __name__ == '__main__':
    main()


import numpy as np

from .mcda_method import MCDA_method
from ..normalizations import linear_normalization



class SAW(MCDA_method):
    def __init__(self, normalization_method = linear_normalization):
        """
        Create the SAW method object and select normalization method `normalization_method`.

        Parameters
        -----------
            normalization_method : function
                method for decision matrix normalization chosen from `normalizations`
        """
        
        self.normalization_method = normalization_method

    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` with m alternatives in rows and 
        n criteria in columns using criteria `weights` and criteria `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 
        
        Examples
        ---------
        >>> saw = SAW(normalization_method = minmax_normalization)
        >>> pref = saw(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """

        SAW._verify_input_data(matrix, weights, types)
        return SAW._saw(matrix, weights, types, self.normalization_method)


    @staticmethod
    def _saw(matrix, weights, types, normalization_method):
        # Normalize matrix using chosen normalization (for example linear normalization)
        norm_matrix = normalization_method(matrix, types)

        # Multiply all rows of normalized matrix by weights
        weighted_matrix = norm_matrix * weights

        # Aggregate and return scores
        return np.sum(weighted_matrix, axis = 1)


import numpy as np

from .mcda_method import MCDA_method
from ..normalizations import sum_normalization



class ARAS(MCDA_method):
    def __init__(self, normalization_method = sum_normalization):
        """
        Create the ARAS method object
        """
        self.normalization_method = normalization_method

    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        --------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ----------
        >>> aras = ARAS()
        >>> pref = aras(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        
        ARAS._verify_input_data(matrix, weights, types)
        return ARAS._aras(matrix, weights, types, self.normalization_method)

    @staticmethod
    def _aras(matrix, weights, types, normalization_method):
        # Create optimal alternative
        A0 = np.zeros(matrix.shape[1])
        A0[types == 1] = np.max(matrix[:, types == 1], axis = 0)
        A0[types == -1] = np.min(matrix[:, types == -1], axis = 0)
        matrix = np.vstack((A0, matrix))
        # Normalize matrix using the sum normalization method
        norm_matrix = normalization_method(matrix, types)
        # Calculate the weighted normalized decision matrix
        d = norm_matrix * weights
        # Calculate the optimality function for each alternative
        S = np.sum(d, axis = 1)
        # Determine the degree of utility
        U = S / S[0]
        return U[1:]


import numpy as np

from .mcda_method import MCDA_method


class MARCOS(MCDA_method):
    def __init__(self):
        """
        Create the MARCOS method object
        """
        pass

    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.
        
        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.
        
        Returns
        --------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 
        
        Examples
        ----------
        >>> marcos = MARCOS()
        >>> pref = marcos(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        
        MARCOS._verify_input_data(matrix, weights, types)
        return MARCOS._marcos(matrix, weights, types)


    @staticmethod
    def _marcos(matrix, weights, types):
        # Step 1 - define the decision matrix with m alternatives and n criteria
        # a decision matrix is provided as an argument of the MARCOS method
        # Step 2
        # Create the ideal solution AI
        AI = np.zeros(matrix.shape[1])
        # for profit criteria
        AI[types == 1] = np.max(matrix[:, types == 1], axis = 0)
        # for cost criteria
        AI[types == -1] = np.min(matrix[:, types == -1], axis = 0)

        # Create the antiideal solution AAI
        AAI = np.zeros(matrix.shape[1])
        # for profit criteria
        AAI[types == 1] = np.min(matrix[:, types == 1], axis = 0)
        # for cost criteria
        AAI[types == -1] = np.max(matrix[:, types == -1], axis = 0)

        
        e_matrix = np.vstack((AAI, matrix))
        e_matrix = np.vstack((e_matrix, AI))

        # Step 3 normalization of extended matrix
        norm_matrix = np.zeros(e_matrix.shape)
        # for cost criteria
        norm_matrix[:, types == -1] = AI[types == -1] / e_matrix[:, types == -1]

        # for profit criteria
        norm_matrix[:, types == 1] = e_matrix[:, types == 1] / AI[types == 1]

        # Step 4 determination of the weighted matrix
        v = norm_matrix * weights

        # Step 5 Calculation of the utility degree of alternatives Ki
        Si = np.sum(v, axis = 1)
        Ki_minus = Si / np.sum(v[0, :])
        Ki_plus = Si / np.sum(v[-1, :])

        # Step 6 Determination of the utility function of alternatives f(Ki).
        f_Ki_minus = Ki_plus / (Ki_plus + Ki_minus)
        f_Ki_plus = Ki_minus / (Ki_plus + Ki_minus)
        f_Ki = (Ki_plus + Ki_minus) / (1 + ((1 - f_Ki_plus) / f_Ki_plus) + ((1 - f_Ki_minus) / f_Ki_minus))

        # Step 7 Ranking the alternatives. 
        # Ranking of the alternatives is based on the final values of utility functions. 
        # It is desirable that an alternative has the highest possible value of 
        # the utility function

        # return utility function values for each alternative
        return f_Ki[1:-1]


import itertools
import numpy as np

from .additions import rank_preferences
from .normalizations import multimoora_normalization


# Copeland Method for compromise ranking
def copeland(matrix):
    """Calculate the compromise ranking considering several rankings obtained using different
    methods using the Copeland compromise ranking methodology.

    Parameters
    -----------
        matrix : ndarray
            Two-dimensional matrix containing different rankings in columns.

    Returns
    --------
        ndarray
            Vector including compromise ranking.

    Examples
    ----------
    >>> rank = copeland(matrix)
    """

    grade_matrix = matrix.shape[0] - matrix
    wins_score = np.sum(grade_matrix, axis = 1)
    lossess_score = np.sum(wins_score) - wins_score
    rank = rank_preferences(wins_score - lossess_score, reverse = True)
    return rank


# Dominance Directed Graph for compromise ranking
def dominance_directed_graph(matrix):
    """Calculate the compromise ranking considering several rankings obtained using different
    methods using Dominance Directed Graph methodology

    Parameters
    -----------
        matrix : ndarray
            Two-dimensional matrix containing different rankings in columns.

    Returns
    --------
        ndarray
            Vector including compromise ranking.

    Examples
    ----------
    >>> rank = dominance_directed_graph(matrix)
    """

    m, n = matrix.shape
    A = np.zeros((m, m))
    for j, i in itertools.product(range(n), range(m)):
        ind_better = np.where(matrix[i, j] < matrix[:, j])[0]
        A[i, j] += len(ind_better)
    AS = np.sum(A, axis = 1)
    ASR = rank_preferences(AS, reverse = True)
    return ASR


# Rank Position Method for compromise ranking
def rank_position_method(matrix):
    """Calculate the compromise ranking considering several rankings obtained using different
    methods using Rank Position Method

    Parameters
    -----------
        matrix : ndarray
            Two-dimensional matrix containing different rankings in columns.

    Returns
    --------
        ndarray
            Vector including compromise ranking.

    Examples
    ---------
    >>> rank = rank_position_method(matrix)
    """

    A = 1 / (np.sum((1 / matrix), axis = 1))
    RPM = rank_preferences(A, reverse = False)
    return RPM


# Improved Borda Rule method for compromise for MULTIMOORA
def improved_borda_rule(prefs, ranks):
    """Calculate the compromise ranking considering several rankings obtained using different
    methods using Improved Borda rule methodology

    Parameters
    -----------
        prefs : ndarray
            Two-dimensional matrix containing preferences calculated by different methods in columns.

        ranks : ndarray
            Two-dimensional matrix containing rankings determined by different methods in columns.

    Returns
    --------
        ndarray
            Vector including compromise ranking.

    Examples
    ----------
    >>> rank = improved_borda_rule(prefs, ranks)
    """

    m, n = ranks.shape
    nprefs = multimoora_normalization(prefs, np.ones(n))
    A = (nprefs[:, 0] * ((m - ranks[:, 0] + 1) / (m * (m + 1) / 2))) - \
    (nprefs[:, 1] * (ranks[:, 1] / (m * (m + 1) / 2))) + \
    (nprefs[:, 2] * ((m - ranks[:, 2] + 1) / (m * (m + 1) / 2)))
    IMB = rank_preferences(A, reverse = True)
    return IMB


import unittest
import numpy as np

from pyrepo_mcda.mcda_methods import CODAS
from pyrepo_mcda.mcda_methods import TOPSIS
from pyrepo_mcda.mcda_methods import WASPAS
from pyrepo_mcda.mcda_methods import VIKOR
from pyrepo_mcda.mcda_methods import SPOTIS
from pyrepo_mcda.mcda_methods import EDAS
from pyrepo_mcda.mcda_methods import MABAC
from pyrepo_mcda.mcda_methods import MULTIMOORA
from pyrepo_mcda.mcda_methods import MULTIMOORA_RS as MOORA

from pyrepo_mcda.mcda_methods.multimoora import MULTIMOORA_RS, MULTIMOORA_RP, MULTIMOORA_FMF

from pyrepo_mcda import distance_metrics as dists
from pyrepo_mcda import correlations as corrs
from pyrepo_mcda import normalizations as norms
from pyrepo_mcda import weighting_methods as mcda_weights
from pyrepo_mcda import compromise_rankings as compromises
from pyrepo_mcda.additions import rank_preferences



# Test for the VIKOR method
class Test_VIKOR(unittest.TestCase):

    def test_vikor(self):
        """Test based on paper Papathanasiou, J., & Ploskas, N. (2018). Vikor. In Multiple Criteria Decision Aid 
        (pp. 31-55). Springer, Cham."""

        matrix = np.array([[8, 7, 2, 1],
        [5, 3, 7, 5],
        [7, 5, 6, 4],
        [9, 9, 7, 3],
        [11, 10, 3, 7],
        [6, 9, 5, 4]])

        weights = np.array([0.4, 0.3, 0.1, 0.2])
        types = np.array([1, 1, 1, 1])

        method = VIKOR(v = 0.625)
        test_result = method(matrix, weights, types)
        real_result = np.array([0.640, 1.000, 0.693, 0.271, 0.000, 0.694])

        self.assertEqual(list(np.round(test_result, 3)), list(real_result))


# Test for the TOPSIS method
class Test_TOPSIS(unittest.TestCase):

    def test_topsis(self):
        """Test based on paper Ersoy, Y. (2021). Equipment selection for an e-commerce company using entropy-based topsis, 
        edas and codas methods during the COVID-19. LogForum, 17(3)."""

        matrix = np.array([[256, 8, 41, 1.6, 1.77, 7347.16],
        [256, 8, 32, 1.0, 1.8, 6919.99],
        [256, 8, 53, 1.6, 1.9, 8400],
        [256, 8, 41, 1.0, 1.75, 6808.9],
        [512, 8, 35, 1.6, 1.7, 8479.99],
        [256, 4, 35, 1.6, 1.7, 7499.99]])

        weights = np.array([0.405, 0.221, 0.134, 0.199, 0.007, 0.034])
        types = np.array([1, 1, 1, 1, -1, -1])

        method = TOPSIS(normalization_method=norms.vector_normalization, distance_metric=dists.euclidean)
        test_result = method(matrix, weights, types)
        real_result = np.array([0.308, 0.254, 0.327, 0.263, 0.857, 0.192])

        self.assertEqual(list(np.round(test_result, 2)), list(np.round(real_result, 2)))


# Test for the CODAS method
class Test_CODAS(unittest.TestCase):

    def test_codas(self):
        """Test based on paper Badi, I., Abdulshahed, A. M., & Shetwan, A. (2018). A case study of supplier selection 
        for a steelmaking company in Libya by using the Combinative Distance-based ASsessment (CODAS) 
        model. Decision Making: Applications in Management and Engineering, 1(1), 1-12."""

        matrix = np.array([[45, 3600, 45, 0.9],
        [25, 3800, 60, 0.8],
        [23, 3100, 35, 0.9],
        [14, 3400, 50, 0.7],
        [15, 3300, 40, 0.8],
        [28, 3000, 30, 0.6]])

        weights = np.array([0.2857, 0.3036, 0.2321, 0.1786])
        types = np.array([1, -1, 1, 1])

        method = CODAS(normalization_method = norms.linear_normalization, distance_metric = dists.euclidean, tau = 0.02)
        test_result = method(matrix, weights, types)
        real_result = np.array([1.3914, 0.3411, -0.2170, -0.5381, -0.7292, -0.2481])
        
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))

    def test_codas2(self):
        """Test based on paper Badi, I., Ballem, M., & Shetwan, A. (2018). Site selection of desalination plant in 
        Libya by using Combinative Distance-based Assessment (CODAS) Method. International Journal 
        for Quality Research, 12(3), 609."""

        matrix = np.array([[8, 8, 10, 9, 5],
        [8, 9, 9, 9, 8],
        [9, 9, 7, 8, 6],
        [8, 8, 7, 8, 9],
        [9, 8, 7, 7, 4]])

        weights = np.array([0.19, 0.26, 0.24, 0.17, 0.14])
        types = np.array([-1, 1, 1, 1, -1])

        method = CODAS(normalization_method = norms.linear_normalization, distance_metric = dists.euclidean, tau = 0.02)
        test_result = method(matrix, weights, types)
        real_result = np.array([0.4463, 0.1658, -0.2544, -0.4618, 0.1041])
        
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))


# Test for the WASPAS method
class Test_WASPAS(unittest.TestCase):

    def test_waspas(self):
        """Test based on paper Tuş, A., & Aytaç Adalı, E. (2019). The new combination with CRITIC and WASPAS methods 
        for the time and attendance software selection problem. Opsearch, 56(2), 528-538."""

        matrix = np.array([[5000, 3, 3, 4, 3, 2],
        [680, 5, 3, 2, 2, 1],
        [2000, 3, 2, 3, 4, 3],
        [600, 4, 3, 1, 2, 2],
        [800, 2, 4, 3, 3, 4]])

        types = np.array([-1, 1, 1, 1, 1, 1])
        weights = mcda_weights.critic_weighting(matrix)

        method = WASPAS(normalization_method=norms.linear_normalization, lambda_param=0.5)
        test_result = method(matrix, weights, types)
        real_result = np.array([0.5622, 0.6575, 0.6192, 0.6409, 0.722])
        
        self.assertEqual(list(np.round(test_result, 3)), list(np.round(real_result, 3)))


# Test for the SPOTIS method
class Test_SPOTIS(unittest.TestCase):

    def test_spotis(self):
        """Test based on paper Dezert, J., Tchamova, A., Han, D., & Tacnet, J. M. (2020, July). 
        The SPOTIS rank reversal free method for multi-criteria decision-making support. 
        In 2020 IEEE 23rd International Conference on Information Fusion (FUSION) (pp. 1-8). 
        IEEE."""

        matrix = np.array([[15000, 4.3, 99, 42, 737],
                 [15290, 5.0, 116, 42, 892],
                 [15350, 5.0, 114, 45, 952],
                 [15490, 5.3, 123, 45, 1120]])

        weights = np.array([0.2941, 0.2353, 0.2353, 0.0588, 0.1765])
        types = np.array([-1, -1, -1, 1, 1])
        bounds_min = np.array([14000, 3, 80, 35, 650])
        bounds_max = np.array([16000, 8, 140, 60, 1300])
        bounds = np.vstack((bounds_min, bounds_max))


        method = SPOTIS()
        test_result = method(matrix, weights, types, bounds)
        
        real_result = np.array([0.4779, 0.5781, 0.5558, 0.5801])
        self.assertEqual(list(np.round(test_result, 3)), list(np.round(real_result, 3)))


# Test for the EDAS method
class Test_EDAS(unittest.TestCase):

    def test_edas(self):
        """Test based on paper Ersoy, Y. (2021). Equipment selection for an e-commerce company using entropy-based topsis, 
        edas and codas methods during the COVID-19. LogForum, 17(3)."""

        matrix = np.array([[256, 8, 41, 1.6, 1.77, 7347.16],
        [256, 8, 32, 1.0, 1.8, 6919.99],
        [256, 8, 53, 1.6, 1.9, 8400],
        [256, 8, 41, 1.0, 1.75, 6808.9],
        [512, 8, 35, 1.6, 1.7, 8479.99],
        [256, 4, 35, 1.6, 1.7, 7499.99]])

        weights = np.array([0.405, 0.221, 0.134, 0.199, 0.007, 0.034])
        types = np.array([1, 1, 1, 1, -1, -1])

        method = EDAS()
        test_result = method(matrix, weights, types)
        real_result = np.array([0.414, 0.130, 0.461, 0.212, 0.944, 0.043])
        self.assertEqual(list(np.round(test_result, 3)), list(np.round(real_result, 3)))


# Test for the MABAC method
class Test_MABAC(unittest.TestCase):

    def test_mabac(self):
        """Test based on paper Isik, O., Aydin, Y., & Kosaroglu, S. M. (2020). The assessment of the logistics 
        Performance Index of CEE Countries with the New Combination of SV and MABAC Methods. 
        LogForum, 16(4)."""

        matrix = np.array([[2.937588, 2.762986, 3.233723, 2.881315, 3.015289, 3.313491],
        [2.978555, 3.012820, 2.929487, 3.096154, 3.012820, 3.593939],
        [3.286673, 3.464600, 3.746009, 3.715632, 3.703427, 4.133620],
        [3.322037, 3.098638, 3.262154, 3.147851, 3.206675, 3.798684],
        [3.354866, 3.270945, 3.221880, 3.213207, 3.670508, 3.785941],
        [2.796570, 2.983000, 2.744904, 2.692550, 2.787563, 2.878851],
        [2.846491, 2.729618, 2.789990, 2.955624, 3.123323, 3.646595],
        [3.253458, 3.208902, 3.678499, 3.580044, 3.505663, 3.954262],
        [2.580718, 2.906903, 3.176497, 3.073653, 3.264727, 3.681887],
        [2.789011, 3.000000, 3.101099, 3.139194, 2.985348, 3.139194],
        [3.418681, 3.261905, 3.187912, 3.052381, 3.266667, 3.695238]])

        weights = np.array([0.171761, 0.105975, 0.191793, 0.168824, 0.161768, 0.199880])
        types = np.array([1, 1, 1, 1, 1, 1])

        method = MABAC(normalization_method=norms.minmax_normalization)
        test_result = method(matrix, weights, types)
        real_result = np.array([-0.155317, -0.089493, 0.505407, 0.132405, 0.246943, -0.386756, 
        -0.179406, 0.362921, -0.084198, -0.167505, 0.139895])
        
        self.assertEqual(list(np.round(test_result, 4)), list(np.round(real_result, 4)))


# Test for the MULTIMOORA method
class Test_MULTIMOORA(unittest.TestCase):

    def test_multimoora(self):
        """Test based on paper Karabasevic, D., Stanujkic, D., Urosevic, S., & Maksimovic, M. (2015). Selection of 
        candidates in the mining industry based on the application of the SWARA and the MULTIMOORA 
        methods. Acta Montanistica Slovaca, 20(2)."""

        matrix = np.array([[4, 3, 3, 4, 3, 2, 4],
        [3, 3, 4, 3, 5, 4, 4],
        [5, 4, 4, 5, 5, 5, 4]])

        weights = np.array([0.215, 0.215, 0.159, 0.133, 0.102, 0.102, 0.073])
        types = np.array([1, 1, 1, 1, 1, 1, 1])

        # MULTIMOORA RS
        method = MULTIMOORA_RS()
        test_result = method(matrix, weights, types)
        real_result = np.array([0.494, 0.527, 0.677])

        self.assertEqual(list(np.round(test_result, 3)), list(np.round(real_result, 3)))

        # MULTIMOORA RP
        method = MULTIMOORA_RP()
        test_result = method(matrix, weights, types)
        real_result = np.array([0.046, 0.061, 0.000])
        
        self.assertEqual(list(np.round(test_result, 3)), list(np.round(real_result, 3)))
        # MULTIMOORA FMF
        method = MULTIMOORA_FMF()
        test_result = method(matrix, weights, types)
        real_result = np.array([3.876e-09, 9.689e-09, 4.486e-08])
        
        self.assertEqual(list(np.round(test_result, 9)), list(np.round(real_result, 9)))

        # MULTIMOORA and `dominance_directed_graph` as compromise ranking
        mmoora = MULTIMOORA(compromise_rank_method=compromises.dominance_directed_graph)
        test_result = mmoora(matrix, weights, types)
        real_result = np.array([3, 2, 1])
        
        self.assertEqual(list(test_result), list(real_result))


# Test for the MOORA method
class Test_MOORA(unittest.TestCase):

    def test_moora(self):
        """Test based on paper Karabasevic, D., Stanujkic, D., Urosevic, S., & Maksimovic, M. (2015). Selection of 
        candidates in the mining industry based on the application of the SWARA and the MULTIMOORA 
        methods. Acta Montanistica Slovaca, 20(2)."""

        matrix = np.array([[4, 3, 3, 4, 3, 2, 4],
        [3, 3, 4, 3, 5, 4, 4],
        [5, 4, 4, 5, 5, 5, 4]])

        weights = np.array([0.215, 0.215, 0.159, 0.133, 0.102, 0.102, 0.073])
        types = np.array([1, 1, 1, 1, 1, 1, 1])

        method = MOORA()
        test_result = method(matrix, weights, types)
        real_result = np.array([0.494, 0.527, 0.677])

        self.assertEqual(list(np.round(test_result, 3)), list(np.round(real_result, 3)))


# Test for rank preferences
class Test_Rank_preferences(unittest.TestCase):

    def test_rank_preferences(self):
        """Test based on paper Test based on paper Papathanasiou, J., & Ploskas, N. (2018). Vikor. In Multiple Criteria Decision Aid 
        (pp. 31-55). Springer, Cham."""

        pref = np.array([0.640, 1.000, 0.693, 0.271, 0.000, 0.694])
        test_result =rank_preferences(pref , reverse = False)
        real_result = np.array([3, 6, 4, 2, 1, 5])
        self.assertEqual(list(test_result), list(real_result))


def main():
    test_vikor = Test_VIKOR()
    test_vikor.test_vikor()

    test_rank_preferences = Test_Rank_preferences()
    test_rank_preferences.test_rank_preferences()

    test_topsis = Test_TOPSIS()
    test_topsis.test_topsis()

    test_codas = Test_CODAS()
    test_codas.test_codas()
    test_codas.test_codas2()

    test_waspas = Test_WASPAS()
    test_waspas.test_waspas()

    test_spotis = Test_SPOTIS()
    test_spotis.test_spotis()

    test_edas = Test_EDAS()
    test_edas.test_edas()

    test_mabac = Test_MABAC()
    test_mabac.test_mabac()

    test_multimoora = Test_MULTIMOORA()
    test_multimoora.test_multimoora()

    test_moora = Test_MOORA()
    test_moora.test_moora()


if __name__ == '__main__':
    main()


import numpy as np
from ..normalizations import sum_normalization
from .mcda_method import MCDA_method


class COPRAS(MCDA_method):
    def __init__(self, normalization_method = sum_normalization):
        """
        Create the COPRAS method object and select normalization method `normalization_method`

        Parameters
        -----------
            normalization_method : function
                method for decision matrix normalization chosen from `normalizations`
        """
        self.normalization_method = normalization_method

    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        --------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ----------
        >>> copras = COPRAS(normalization_method = sum_normalization)
        >>> pref = copras(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """

        COPRAS._verify_input_data(matrix, weights, types)
        return COPRAS._copras(matrix, weights, types, self.normalization_method)

    @staticmethod
    def _copras(matrix, weights, types, normalization_method):
        # Normalize matrix as for profit criteria using chosen normalization method.
        # norm_matrix = matrix/np.sum(matrix, axis = 0)
        norm_matrix = normalization_method(matrix, np.ones(len(weights)))
        # Multiply all values in the normalized matrix by weights.
        d = norm_matrix * weights
        # Calculate the sums of weighted normalized outcomes for profit criteria.
        Sp = np.sum(d[:, types == 1], axis = 1)
        # Calculate the sums of weighted normalized outcomes for cost criteria.
        Sm = np.sum(d[:, types == -1], axis = 1)
        # Calculate the relative priority Q of evaluated options.
        Q = Sp + ((np.sum(Sm))/(Sm * np.sum(1 / Sm)))
        # Calculate the quantitive utility value for each of the evaluated options.
        U = Q / np.max(Q)
        return U





import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import copy

from pyrepo_mcda.mcda_methods import TOPSIS, VIKOR, CODAS, SPOTIS, PROMETHEE_II, COPRAS
from pyrepo_mcda.additions import rank_preferences
from pyrepo_mcda import weighting_methods as mcda_weights
from pyrepo_mcda import correlations as corrs
from pyrepo_mcda import normalizations as norms
from pyrepo_mcda import distance_metrics as dists

# Create dictionary class to create a data frame with correlation values
class Create_dictionary(dict):
  
    # __init__ function
    def __init__(self):
        self = dict()
          
    # Function to add key:value
    def add(self, key, value):
        self[key] = value

# bar (column) chart
def plot_barplot(df_plot, stacked = False):
    """
    Visualization method to display column chart of alternatives rankings obtained with 
    different methods.

    Parameters
    ----------
        df_plot : DataFrame
            DataFrame containing rankings of alternatives obtained with different methods.
            The particular rankings are included in subsequent columns of DataFrame.

        stacked : Boolean
            Variable denoting if the chart is to be stacked or not.

    Examples
    ----------
    >>> plot_barplot(df_plot)
    """

    ax = df_plot.plot(kind='bar', width = 0.8, stacked=stacked, edgecolor = 'black', figsize = (9,4))
    if stacked == False:
        step = 1
        list_rank = np.arange(1, len(df_plot) + 1, step)
        ax.set_yticks(list_rank)
        ax.set_ylim(0, len(df_plot) + 1)

    ax.set_xticklabels(df_plot.index, rotation = 'horizontal')
    ax.tick_params(axis = 'both', labelsize = 12)
    y_ticks = ax.yaxis.get_major_ticks()
    
    ncol = df_plot.shape[1]
    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc='lower left',
    ncol = ncol, mode="expand", borderaxespad=0., edgecolor = 'black', fontsize = 12)

    ax.grid(True)
    ax.set_axisbelow(True)
    plt.tight_layout()
    plt.show()


# plot line chart
def plot_lineplot(df_plot):
    """
    Visualization method to display line chart of alternatives rankings.

    Parameters
    ----------
        df_plot : DataFrame
            DataFrame containing rankings of alternatives obtained with different weight of
            selected criterion. The particular rankings are contained in subsequent columns of 
            DataFrame.
        
    Examples
    ----------
    >>> plot_lineplot(df_plot)
    """

    plt.figure(figsize = (9, 4))
    for j in range(df_plot.shape[0]):
        
        plt.plot(df_plot.iloc[j, :], '-o', linewidth = 2)
        ax = plt.gca()
        y_min, y_max = ax.get_ylim()
        x_min, x_max = ax.get_xlim()
        plt.annotate(df_plot.index[j], (x_max, df_plot.iloc[j, -1]),
                        fontsize = 12, style='italic',
                        horizontalalignment='left')

    plt.xlabel("Method", fontsize = 12)
    plt.ylabel("Rank", fontsize = 12)
    plt.yticks(np.arange(0, df_plot.shape[0]) + 1, fontsize = 12)
    plt.xticks(fontsize = 12)
    plt.grid(True)
    plt.tight_layout()
    plt.show()


# plot radar chart
def plot_radar(data):
    """
    Visualization method to display rankings of alternatives obtained with different methods
    on the radar chart.

    Parameters
    -----------
        data : DataFrame
            DataFrame containing containing rankings of alternatives obtained with different 
            methods. The particular rankings are contained in subsequent columns of DataFrame.
        
    Examples
    ----------
    >>> plot_radar(data)
    """

    fig=plt.figure()
    ax = fig.add_subplot(111, polar = True)

    for col in list(data.columns):
        labels=np.array(list(data.index))
        stats = data.loc[labels, col].values

        angles=np.linspace(0, 2*np.pi, len(labels), endpoint=False)
        # close the plot
        stats=np.concatenate((stats,[stats[0]]))
        angles=np.concatenate((angles,[angles[0]]))
    
        lista = list(data.index)
        lista.append(data.index[0])
        labels=np.array(lista)

        ax.plot(angles, stats, '-o', linewidth=2)
    
    ax.set_thetagrids(angles * 180/np.pi, labels)
    ax.set_rgrids(np.arange(1, data.shape[0] + 1, 1))
    ax.grid(True)
    ax.set_axisbelow(True)
    # plt.legend(data.columns, bbox_to_anchor=(1.0, 0.95, 0.4, 0.2), loc='upper left')
    if data.shape[1] % 2 == 0:
        ncol = data.shape[1] // 2
    else:
        ncol = data.shape[1] // 2 + 1
    plt.legend(data.columns, bbox_to_anchor=(-0.1, 1.1, 1.2, .102), loc='lower left',
    ncol = ncol, mode="expand", borderaxespad=0., edgecolor = 'black', fontsize = 12)
    plt.tight_layout()
    plt.show()


# plot scatter chart
def plot_scatter(data, model_compare):
    """
    Display scatter plot comparing real and predicted ranking.

    Parameters
    -----------
        data: dataframe
        model_compare : list[list] with two columns including compared rankings

    Examples
    ----------
    >>> plot_scatter(data)
    """
    
    list_rank = np.arange(1, len(data) + 1, 1)
    list_alt_names = data.index
    for it, el in enumerate(model_compare):
        
        xx = [min(data[el[0]]), max(data[el[0]])]
        yy = [min(data[el[1]]), max(data[el[1]])]

        fig, ax = plt.subplots(figsize=(7, 7))
        ax.plot(xx, yy, linestyle = '--', zorder = 1)

        ax.scatter(data[el[0]], data[el[1]], marker = 'o', color = 'royalblue', zorder = 2)
        for i, txt in enumerate(list_alt_names):
            ax.annotate(txt, (data[el[0]][i], data[el[1]][i]), fontsize = 14, style='italic',
                         verticalalignment='bottom', horizontalalignment='right')

        ax.set_xlabel(el[0], fontsize = 16)
        ax.set_ylabel(el[1], fontsize = 16)
        ax.tick_params(axis='both', labelsize=16)
        ax.set_xticks(list_rank)
        ax.set_yticks(list_rank)

        x_ticks = ax.xaxis.get_major_ticks()
        y_ticks = ax.yaxis.get_major_ticks()

        ax.set_xlim(0.5, len(data) + 0.5)
        ax.set_ylim(0.5, len(data) + 0.5)

        ax.grid(True, linestyle = '--')
        ax.set_axisbelow(True)
    
        plt.tight_layout()
        plt.show()


# boxplot chart for visualization of distribution
def plot_boxplot(data):
    """
    Display boxplot showing distribution of preference values determined with different methods.
    Parameters
    ----------
        data : dataframe
            dataframe with correlation values between compared rankings
        
    Examples
    ----------
    >>> plot_boxplot(data, title)
    """
    
    df_melted = pd.melt(data)
    plt.figure(figsize = (7, 4))
    ax = sns.boxplot(x = 'variable', y = 'value', data = df_melted, width = 0.4)
    ax.set_xlabel('', fontsize = 12)
    ax.set_ylabel('', fontsize = 12)
    ax.grid(True, linestyle = '--')
    ax.set_axisbelow(True)
    plt.tight_layout()
    plt.show()


def plot_boxplot_1(data):
    """
    Display boxplot showing distribution of criteria weights determined with different methods.

    Parameters
    ----------
    data : dataframe
        dataframe with correlation values between compared rankings
    """
    
    plt.figure(figsize = (7, 4))
    
    ax = data.boxplot()
    ax.grid(True, linestyle = '--')
    ax.set_axisbelow(True)
    plt.tight_layout()
    plt.show()


def plot_boxplot_2(data):
    """
    Display boxplot showing distribution of criteria weights determined with different methods.

    Parameters
    ----------
    data : dataframe
        dataframe with correlation values between compared rankings
    """
    
    plt.figure(figsize = (7, 4))
    
    plt.boxplot(data)
    plt.xticks(np.arange(1, data.shape[1] + 1), data.columns)
    plt.grid(True, linestyle = '--')
    plt.tight_layout()
    plt.show()


def plot_boxenplot(data):
    ax = sns.boxenplot(data = data)
    ax.grid(True)
    ax.set_axisbelow(True)
    plt.tight_layout()
    plt.show()


# heat maps with correlations
def draw_heatmap(df_new_heatmap):
    """
    Visualization method to display heatmap with correlations of compared rankings generated using different methods
    
    Parameters
    ----------
        data : DataFrame
            DataFrame with correlation values between compared rankings
        
    Examples
    ---------
    >>> draw_heatmap(df_new_heatmap)
    """
    plt.figure(figsize = (8, 5))
    sns.set(font_scale = 1.2)
    heatmap = sns.heatmap(df_new_heatmap, annot=True, fmt=".2f", cmap="PuBu",
                          linewidth=0.5, linecolor='w')
    plt.yticks(va="center")
    plt.tight_layout()
    plt.show()



# TODO add violin plot z losowaniem czegos na przyklad ARAS COPRAS i normalizations
# to co mialo byc na jakas konfe

def main():

    # load decision matrix from csv file

    data = pd.read_csv('dataset_cars.csv', index_col = 'Ai')

    df_data = data.iloc[:len(data) - 1, :]
    types = data.iloc[len(data) - 1, :].to_numpy()
    matrix = df_data.to_numpy()
    weights = mcda_weights.entropy_weighting(matrix)

    list_alt_names = [r'$A_{' + str(i) + '}$' for i in range(1, df_data.shape[0] + 1)]
    cols = [r'$C_{' + str(j) + '}$' for j in range(1, data.shape[1] + 1)]

    # Calculate rankings of different MCDA methods
    rank_results = pd.DataFrame(index=list_alt_names)

    # TOPSIS
    topsis = TOPSIS(normalization_method = norms.minmax_normalization, distance_metric = dists.euclidean)
    pref = topsis(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['TOPSIS'] = rank

    # VIKOR
    vikor = VIKOR(normalization_method = norms.minmax_normalization)
    pref = vikor(matrix, weights, types)
    rank = rank_preferences(pref, reverse = False)
    rank_results['VIKOR'] = rank

    # CODAS
    codas = CODAS(normalization_method = norms.linear_normalization, distance_metric = dists.euclidean, tau = 0.02)
    pref = codas(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['CODAS'] = rank

    # SPOTIS
    bounds_min = np.amin(matrix, axis = 0)
    bounds_max = np.amax(matrix, axis = 0)
    bounds = np.vstack((bounds_min, bounds_max))
    spotis = SPOTIS()
    pref = spotis(matrix, weights, types, bounds)
    rank = rank_preferences(pref, reverse = False)
    rank_results['SPOTIS'] = rank

    # PROMETHEE II
    promethee_II = PROMETHEE_II()
    pref = promethee_II(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['PROMETHEE II'] = rank

    # COPRAS
    copras = COPRAS(normalization_method = norms.sum_normalization)
    pref = copras(matrix, weights, types)
    rank = rank_preferences(pref, reverse=True)
    rank_results['COPRAS'] = rank

    # bar chart
    plot_barplot(rank_results)

    # line chart
    plot_lineplot(rank_results)

    # radar chart
    plot_radar(rank_results)

    # plot scatter chart
    model_compare = []
    names = list(rank_results.columns)
    model_compare = [[names[0], names[1]]]
    plot_scatter(rank_results, model_compare)

    # Criteria weights distribution
    # Create a list with weighting methods that you want to explore
    weighting_methods_set = [
        mcda_weights.equal_weighting,
        mcda_weights.entropy_weighting,
        mcda_weights.critic_weighting,
        mcda_weights.gini_weighting,
        mcda_weights.merec_weighting,
        mcda_weights.stat_var_weighting,
        mcda_weights.idocriw_weighting,
        mcda_weights.angle_weighting,
        mcda_weights.coeff_var_weighting
    ]

    df_weights = pd.DataFrame(index = cols)
    df_preferences = pd.DataFrame(index = list_alt_names)
    df_rankings = pd.DataFrame(index = list_alt_names)

    # Create dataframes for weights, preference function values and rankings determined using different weighting methods
    df_weights = pd.DataFrame(index = cols)
    df_preferences = pd.DataFrame(index = list_alt_names)
    df_rankings = pd.DataFrame(index = list_alt_names)

    # Create the VIKOR method object
    vikor = VIKOR()
    for weight_type in weighting_methods_set:
        
        if weight_type.__name__ in ["cilos_weighting", "idocriw_weighting", "angle_weighting", "merec_weighting"]:
            weights = weight_type(matrix, types)
        else:
            weights = weight_type(matrix)
            
        df_weights[weight_type.__name__[:-10].upper().replace('_', ' ')] = weights
        pref = vikor(matrix, weights, types)
        rank = rank_preferences(pref, reverse = False)
        df_preferences[weight_type.__name__[:-10].upper().replace('_', ' ')] = pref
        df_rankings[weight_type.__name__[:-10].upper().replace('_', ' ')] = rank

    # distribution of criteria weights
    plot_boxplot(df_weights.T)
    plot_boxplot_1(df_weights.T)
    plot_boxplot_2(df_weights.T)
    print(df_weights.T)
    plot_boxenplot(df_weights.T)
    # bar chart of criteria weights values for different weighting methods
    plot_barplot(df_weights.T, stacked = True)

    # Rankings correlations
    results = copy.deepcopy(rank_results)
    method_types = list(results.columns)
    dict_new_heatmap_rw = Create_dictionary()

    for el in method_types:
        dict_new_heatmap_rw.add(el, [])

    # heatmaps for correlations coefficients
    for i, j in [(i, j) for i in method_types[::-1] for j in method_types]:
        dict_new_heatmap_rw[j].append(corrs.weighted_spearman(results[i], results[j]))

    df_new_heatmap_rw = pd.DataFrame(dict_new_heatmap_rw, index = method_types[::-1])
    df_new_heatmap_rw.columns = method_types

    # correlation matrix with rw coefficient
    draw_heatmap(df_new_heatmap_rw)


if __name__ == '__main__':
    main()


import copy
import numpy as np
import pandas as pd
import matplotlib
from tabulate import tabulate

from visualizations import plot_barplot, draw_heatmap, plot_boxplot, plot_lineplot_sensitivity, plot_barplot_sensitivity, plot_radar, plot_boxplot_simulation
from pyrepo_mcda.mcda_methods import CODAS, TOPSIS, WASPAS, VIKOR, SPOTIS, EDAS, MABAC, MULTIMOORA, CRADIS
from pyrepo_mcda.mcda_methods import MULTIMOORA_RS as MOORA
from pyrepo_mcda import distance_metrics as dists
from pyrepo_mcda import correlations as corrs
from pyrepo_mcda import normalizations as norms
from pyrepo_mcda import weighting_methods as mcda_weights
from pyrepo_mcda import compromise_rankings as compromises
from pyrepo_mcda.additions import rank_preferences
from pyrepo_mcda.sensitivity_analysis_weights_percentages import Sensitivity_analysis_weights_percentages
from pyrepo_mcda.sensitivity_analysis_weights_values import Sensitivity_analysis_weights_values


# Create dictionary class
class Create_dictionary(dict):
  
    # __init__ function
    def __init__(self):
        self = dict()
          
    # Function to add key:value
    def add(self, key, value):
        self[key] = value



def main():
    # load name of file with input data
    data = pd.read_csv('data' + '.csv', index_col = 'Ai')
    df_data = data.iloc[:len(data) - 2, :]
    weights = data.iloc[len(data) - 2, :].to_numpy()
    types = data.iloc[len(data) - 1, :].to_numpy()

    matrix = df_data.to_numpy()

    header = [df_data.index.name]
    header = header + list(df_data.columns)
    print(tabulate(df_data, headers = header, tablefmt='github'))
    
    print(types)

    list_alt_names = [r'$A_{' + str(i) + '}$' for i in range(1, df_data.shape[0] + 1)]
    list_crit_names = [r'$C_{' + str(i) + '}$' for i in range(1, df_data.shape[1] + 1)]
    
    matrix = df_data.to_numpy()

    print(weights)

    # Distance metrics used with TOPSIS
    # Exploration of results for different chosen distance metrics
    distance_metrics = [
        dists.euclidean,
        dists.manhattan,
        dists.hausdorff,
        dists.chebyshev,
        dists.bray_curtis,
        dists.canberra,
        dists.lorentzian,
        dists.jaccard
    ]

    # Create dataframes for preference function values and rankings determined using distance metrics
    df_preferences = pd.DataFrame(index = list_alt_names)
    df_rankings = pd.DataFrame(index = list_alt_names)

    for distance_metric in distance_metrics:
        # Create the TOPSIS method object
        topsis = TOPSIS(normalization_method = norms.minmax_normalization, distance_metric = distance_metric)
        pref = topsis(matrix, weights, types)
        rank = rank_preferences(pref, reverse = True)
        df_preferences[distance_metric.__name__.capitalize().replace('_', ' ')] = pref
        df_rankings[distance_metric.__name__.capitalize().replace('_', ' ')] = rank
        

    print(df_rankings)
    # plot box chart of alternatives preference values
    plot_barplot(df_rankings, legend_title = 'Distance metrics')

    # plot box chart of alternatives preference values
    # plot_boxplot(df_preferences.T, 'TOPSIS, preference values distribution')

    # Exploration of different MCDA methods
    rank_results = pd.DataFrame()
    rank_results['Ai'] = list(list_alt_names)

    # TOPSIS
    # TOPSIS preferences must be sorted in descending order
    topsis = TOPSIS(normalization_method = norms.minmax_normalization, distance_metric = dists.euclidean)
    pref = topsis(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['TOPSIS'] = rank


    # CODAS
    # CODAS preferences must be sorted in descending order
    codas = CODAS(normalization_method = norms.linear_normalization, distance_metric = dists.euclidean, tau = 0.02)
    pref = codas(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['CODAS'] = rank


    # VIKOR
    # VIKOR preferences must be sorted in ascending order
    vikor = VIKOR(normalization_method = norms.minmax_normalization)
    pref = vikor(matrix, weights, types)
    rank = rank_preferences(pref, reverse = False)
    rank_results['VIKOR'] = rank


    # SPOTIS
    # SPOTIS preferences must be sorted in ascending order
    bounds_min = np.amin(matrix, axis = 0)
    bounds_max = np.amax(matrix, axis = 0)
    bounds = np.vstack((bounds_min, bounds_max))
    spotis = SPOTIS()
    pref = spotis(matrix, weights, types, bounds)
    rank = rank_preferences(pref, reverse = False)
    rank_results['SPOTIS'] = rank


    # EDAS 
    # EDAS preferences must be sorted in descending order
    edas = EDAS()
    pref = edas(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['EDAS'] = rank


    # MABAC
    # MABAC preferences must be sorted in descending order
    mabac = MABAC(normalization_method = norms.minmax_normalization)
    pref = mabac(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['MABAC'] = rank


    # MULTIMOORA
    # MULTIMOORA method returns rank
    multimoora = MULTIMOORA()
    rank = multimoora(matrix, weights, types)
    rank_results['MMOORA'] = rank


    # WASPAS
    # WASPAS preferences must be sorted descending order
    waspas = WASPAS(normalization_method = norms.linear_normalization, lambda_param = 0.5)
    pref = waspas(matrix, weights, types)
    rank = rank_preferences(pref, reverse = True)
    rank_results['WASPAS'] = rank


    rank_results = rank_results.set_index('Ai')
    header = [rank_results.index.name]
    header = header + list(rank_results.columns)
    print(tabulate(rank_results, headers = header, tablefmt='orgtbl'))

    # compromise ranking prepared with the Copeland compromise ranking

    compromise_ranking = compromises.copeland(rank_results)

    rank_results_final = copy.deepcopy(rank_results)
    rank_results_final['Compromise'] = compromise_ranking
    header = [rank_results_final.index.name]
    header = header + list(rank_results_final.columns)
    print(tabulate(rank_results_final, headers = header, tablefmt='github'))
    rank_results_final.to_csv('./results/' + 'results.csv')

    
    # visualization of rankings provided by different MCDA methods
    # barplot
    df_plot = copy.deepcopy(rank_results)
    plot_barplot(df_plot, legend_title='MCDA methods')
    
    
    #
    # correlations heatmaps
    data = copy.deepcopy(rank_results_final)
    method_types = list(data.columns)

    dict_new_heatmap_rw = Create_dictionary()

    for el in method_types:
        dict_new_heatmap_rw.add(el, [])

    dict_new_heatmap_ws = copy.deepcopy(dict_new_heatmap_rw)
    dict_new_heatmap_pearson = copy.deepcopy(dict_new_heatmap_rw)
    dict_new_heatmap_spearman = copy.deepcopy(dict_new_heatmap_rw)

    # heatmaps for correlations coefficients
    for i, j in [(i, j) for i in method_types[::-1] for j in method_types]:
        dict_new_heatmap_rw[j].append(corrs.weighted_spearman(data[i], data[j]))
        dict_new_heatmap_ws[j].append(corrs.WS_coeff(data[i], data[j]))
        dict_new_heatmap_pearson[j].append(corrs.pearson_coeff(data[i], data[j]))
        dict_new_heatmap_spearman[j].append(corrs.spearman(data[i], data[j]))
        

    df_new_heatmap_rw = pd.DataFrame(dict_new_heatmap_rw, index = method_types[::-1])
    df_new_heatmap_rw.columns = method_types

    df_new_heatmap_ws = pd.DataFrame(dict_new_heatmap_ws, index = method_types[::-1])
    df_new_heatmap_ws.columns = method_types

    df_new_heatmap_pearson = pd.DataFrame(dict_new_heatmap_pearson, index = method_types[::-1])
    df_new_heatmap_pearson.columns = method_types

    df_new_heatmap_spearman = pd.DataFrame(dict_new_heatmap_spearman, index = method_types[::-1])
    df_new_heatmap_spearman.columns = method_types
    
    # correlation matrix with rw coefficient
    draw_heatmap(df_new_heatmap_rw, r'$r_w$')

    # correlation matrix with WS coefficient
    draw_heatmap(df_new_heatmap_ws, r'$WS$')

    # correlation matrix with Pearson coefficient
    draw_heatmap(df_new_heatmap_pearson, r'$Pearson$')

    # correlation matrix with Spearman coefficient
    draw_heatmap(df_new_heatmap_spearman, r'$r_s$')
    
    matplotlib.rc_file_defaults()
    print('Sensitivity analysis')
    # Sensitivity analysis - percentage modification of chosen criterion weight
    
    
    # load input vector with percentage values of chosen criterion weights modification for sensitivity analysis
    # percentages = np.arange(0.25, 0.55, 0.1)
    percentages = np.arange(0.05, 0.55, 0.1)
    

    # Create the object of chosen MCDA method
    method = CRADIS(normalization_method=norms.linear_normalization)
    
    # Create the sensitivity analysis method object
    sensitivity_analysis = Sensitivity_analysis_weights_percentages()
    
    # Perform sensitivity analysis with weights percentage modification for chosen criteria
    for j in [0, 1, 2, 3]:
        df_sens = sensitivity_analysis(matrix, weights, types, percentages, method, j, [-1, 1])

        header = [df_sens.index.name]
        header = header + list(df_sens.columns)
        print('Sensitivity analysis for C' + str(j + 1))
        print(tabulate(df_sens, headers = header, tablefmt='github'))

        plot_lineplot_sensitivity(df_sens, method.__class__.__name__, list_crit_names[j], "Weight modification in %", "percentage")
        # plot_barplot_sensitivity(df_sens, method.__class__.__name__, list_crit_names[j], "weight_percentage_bar")
    
    # Perform sensitivity analysis with setting chosen weight value to selected criterion
    # other criteria have equal weight values and all criteria weights sum to 1
    sensitivity_analysis_weights_values = Sensitivity_analysis_weights_values()
    weight_values = np.arange(0.05, 0.95, 0.1)
    for j in [0, 1, 2, 3]:
        df_sens = sensitivity_analysis_weights_values(matrix, weight_values, types, method, j)
        header = [df_sens.index.name]
        header = header + list(df_sens.columns)
        print('Sensitivity analysis for C' + str(j + 1))
        print(tabulate(df_sens, headers = header, tablefmt='github'))
        plot_lineplot_sensitivity(df_sens, method.__class__.__name__, list_crit_names[j], "Weight value", "value")
    # example of bar chart usage for sensitivity analysis
    plot_barplot_sensitivity(df_sens, method.__class__.__name__, list_crit_names[j], "weight_values_bar")
    # example of radar chart usage for sensitivity analysis
    plot_radar(df_sens, list_crit_names[j] + ' weight modification', j)


    # Robustness analysis
    # Determining intervals of alternatives performance values for particular positions in rankings
    # Create object of chosen MCDA method
    topsis = TOPSIS(normalization_method=norms.minmax_normalization, distance_metric=dists.euclidean)

    # Create minimum bounds of criteria performance
    bounds_min = np.amin(matrix, axis = 0)
    # Create maximum bounds of criteria performance
    bounds_max = np.amax(matrix, axis = 0)
    bounds = np.vstack((bounds_min, bounds_max))

    # Create ideal Solution `isp`
    isp = np.zeros(matrix.shape[1])
    isp[types == 1] = bounds[1, types == 1]
    isp[types == -1] = bounds[0, types == -1]

    # Create anti-Ideal Solution `asp`
    asp = np.zeros(matrix.shape[1])
    asp[types == 1] = bounds[0, types == 1]
    asp[types == -1] = bounds[1, types == -1]

    # Create dictionary with values of stepwise particular criteria performance change
    indexes = {
        0 : 1,
        1 : 10,
        2 : 5,
        3 : 0.1
    }
    
    # Iterate by all criteria
    for j in range(matrix.shape[1]):
        change_val = indexes[j]
        # dictionary for collecting ranking values after modification of performance values
        dict_results_sim = {
            'Rank' : [],
            'Performance' : [],
            'Alternative' : []
            }
        # Iterate by all Alternatives
        for i in range(matrix.shape[0]):
            vec = np.arange(asp[j], isp[j] + types[j] * change_val, types[j] * change_val)
            # Iterate by all performance values of chosen criterion
            for v in vec:
                new_matrix = copy.deepcopy(matrix)
                new_matrix[i, j] = v
                pref = topsis(new_matrix, weights, types)
                rank = rank_preferences(pref, reverse = True)
                dict_results_sim['Rank'].append(rank[i])
                dict_results_sim['Performance'].append(v)
                dict_results_sim['Alternative'].append(list_alt_names[i])

        df_results_sim = pd.DataFrame(dict_results_sim)
        df_results_sim.to_csv('results/' + 'robustness_C' + str(j + 1) + '.csv')

        plot_boxplot_simulation(df_results_sim, 'Alternative', 'Performance', 'Rank' , 'Alternative', 'Performance', 'TOPSIS, Criterion ' + list_crit_names[j] + ' performance change', 'robustness_C' + str(j + 1))
    

if __name__ == "__main__":
    main()


import numpy as np
from .mcda_method import MCDA_method


class EDAS(MCDA_method):
    def __init__(self):
        """Create object of the EDAS method"""
        pass


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vevtor with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        --------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ---------
        >>> edas = EDAS()
        >>> pref = edas(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        EDAS._verify_input_data(matrix, weights, types)
        return EDAS._edas(matrix, weights, types)


    def _edas(matrix, weights, types):
        m, n = matrix.shape
        #AV = np.mean(matrix, axis = 0)

        # Calculate the average solution for each criterion
        AV = np.sum(matrix, axis = 0) / m

        # Calculate the Positive Distance (PDA) and Negative Distance (NDA) from average solution
        PDA = np.zeros(matrix.shape)
        NDA = np.zeros(matrix.shape)

        for j in range(0, n):
            if types[j] == 1:
                PDA[:, j] = (matrix[:, j] - AV[j]) / AV[j]
                NDA[:, j] = (AV[j] - matrix[:, j]) / AV[j]
            else:
                PDA[:, j] = (AV[j] - matrix[:, j]) / AV[j]
                NDA[:, j] = (matrix[:, j] - AV[j]) / AV[j]

        PDA[PDA < 0] = 0
        NDA[NDA < 0] = 0

        # Calculate the weighted sum of PDA and NDA for all alternatives
        SP = np.sum(weights * PDA, axis = 1)
        SN = np.sum(weights * NDA, axis = 1)

        # Normalize obtained values
        NSP = SP / np.max(SP)
        NSN = 1 - (SN / np.max(SN))

        # Calculate the appraisal score (AS) for each alternative 
        AS = (NSP + NSN) / 2
        return AS




import copy
import itertools
import numpy as np


# Euclidean distance
def euclidean(A, B):
    """
    Calculate Euclidean distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = euclidean(A, B)
    """

    tmp = np.sqrt(np.sum(np.square(A - B)))
    return tmp

# Manhattan distance
def manhattan(A, B):
    """
    Calculate Manhattan (Taxicab) distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = manhattan(A, B)
    """

    tmp = np.sum(np.abs(A - B))
    return tmp


# function for calculation Hausdorff distance using `hausdorff` function
def hausdorff_distance(A, B):
    min_h = np.inf
    for i, j in itertools.product(range(len(A)), range(len(B))):
        d = euclidean(A[i], B[j])
        if d < min_h:
            min_h = d
            min_ind = j

    max_h = -np.inf
    for i in range(len(A)):
        d = euclidean(A[i], B[min_ind])
        if d > max_h:
            max_h = d

    return max_h


# Hausdorff distance
def hausdorff(A, B):
    """
    Calculate Hausdorff distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = hausdorff(A, B)
    """

    ah = hausdorff_distance(A, B)
    bh = hausdorff_distance(B, A)
    tmp = max(ah, bh)
    return tmp


# correlation distance
def correlation(A, B):
    """
    Calculate Correlation distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = correlation(A, B)
    """

    numerator = np.sum((A - np.mean(A)) * (B - np.mean(B)))
    denominator = np.sqrt(np.sum((A - np.mean(A)) ** 2)) * np.sqrt(np.sum((B - np.mean(B)) ** 2))
    if denominator == 0:
        denominator = 1
    tmp = 1 - (numerator / denominator)
    return tmp


# Chebyshev distance
def chebyshev(A, B):
    """
    Calculate Chebyshev distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = chebyshev(A, B)
    """

    max_h = -np.inf
    for i, j in itertools.product(range(len(A)), range(len(B))):
        d = np.abs(A[i] - B[j])
        if d > max_h:
            max_h = d

    return max_h


# standardized Euclidean distance
def std_euclidean(A, B):
    """
    Calculate Standardized Euclidean distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = std_euclidean(A, B)
    """

    tab_std = np.vstack((A, B))
    stdv = np.sum(np.square(tab_std - np.mean(tab_std, axis = 0)), axis = 0)
    stdv = np.sqrt(stdv / tab_std.shape[0])
    stdv[stdv == 0] = 1
    tmp = np.sum(np.square((A - B) / stdv))
    tmp = np.sqrt(tmp)
    return tmp


# cosine distance
def cosine(A, B):
    """
    Calculate Cosine distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = cosine(A, B)
    """

    numerator = np.sum(A * B)
    denominator = (np.sqrt(np.sum(np.square(A)))) * (np.sqrt(np.sum(np.square(B))))
    if denominator == 0:
        denominator = 1
    tmp = 1 - (numerator / denominator)
    return tmp


# cosine similarity measure
def csm(A, B):
    """
    Calculate Cosine similarity measure of distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ---------
    >>> distance = csm(A, B)
    """

    numerator = np.sum(A * B)
    denominator = (np.sqrt(np.sum(A))) * (np.sqrt(np.sum(B)))
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# squared Euclidean distance
def squared_euclidean(A, B):
    """
    Calculate Squared Euclidean distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = squared_euclidean(A, B)
    """

    tmp = np.sum(np.square(A - B))
    return tmp


# Sorensen or Bray-curtis distance
def bray_curtis(A, B):
    """
    Calculate Bray-Curtis distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = bray_curtis(A, B)
    """

    numerator = np.sum(np.abs(A - B))
    denominator = np.sum(A + B)
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# Canberra distance
def canberra(A, B):
    """
    Calculate Canberra distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = canberra(A, B)
    """

    numerator = np.abs(A - B)
    denominator = A + B
    denominator[denominator == 0] = 1
    tmp = np.sum(numerator / denominator)
    return tmp


# Lorentzian distance
def lorentzian(A, B):
    """
    Calculate Lorentzian distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = lorentzian(A, B)
    """

    tmp = np.sum(np.log(1 + np.abs(A - B)))
    return tmp


# Jaccard distance
def jaccard(A, B):
    """
    Calculate Jaccard distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    -----------
    >>> distance = jaccard(A, B)
    """

    numerator = np.sum(np.square(A - B))
    denominator = np.sum(A ** 2) + np.sum(B ** 2) - np.sum(A * B)
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# Dice distance
def dice(A, B):
    """
    Calculate Dice distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = dice(A, B)
    """

    numerator = np.sum(np.square(A - B))
    denominator = np.sum(A ** 2) + np.sum(B ** 2)
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# Bhattacharyya distance
def bhattacharyya(A, B):
    """
    Calculate Bhattacharyya distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ---------
    >>> distance = bhattacharyya(A, B)
    """

    value = (np.sum(np.sqrt(A * B)))**2
    if value == 0:
        tmp = 0
    else:
        tmp = -np.log(value)
    return tmp


# Hellinger distance
def hellinger(A, B):
    """
    Calculate Hellinger distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    -----------
    >>> distance = hellinger(A, B)
    """

    value = 1 - np.sum(np.sqrt(A * B))
    if value < 0:
        value = 0
    tmp = 2 * np.sqrt(value)
    return tmp


# Matusita distance
def matusita(A, B):
    """
    Calculate Matusita distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = matusita(A, B)
    """

    value = 2 - 2 * (np.sum(np.sqrt(A * B)))
    if value < 0:
        value = 0
    tmp = np.sqrt(value)
    return tmp


# squared-chord distance
def squared_chord(A, B):
    """
    Calculate Squared-Chord distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ---------
    >>> distance = squared_chord(A, B)
    """

    tmp = np.sum(np.square(np.sqrt(A) - np.sqrt(B)))
    return tmp


# Pearson chi-square distance
def pearson_chi_square(A, B):
    """
    Calculate Pearson Chi Square distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ---------
    >>> distance = pearson_chi_square(A, B)
    """

    numerator = np.square(A - B)
    denominator = copy.deepcopy(B)
    denominator[denominator == 0] = 1
    tmp = np.sum(numerator / denominator)
    return tmp

# squared chi-square distance
def squared_chi_square(A, B):
    """
    Calculate Squared Chi Sqaure distance between two vectors `A` and `B`.

    Parameters
    -----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    --------
        float
            distance value between two vectors

    Examples
    ----------
    >>> distance = squared_chi_square(A, B)
    """
    
    numerator = np.square(A - B)
    denominator = A + B
    denominator[denominator == 0] = 1
    tmp = np.sum(numerator / denominator)
    return tmp


from abc import ABC
import numpy as np

class MCDA_method(ABC):

    def __call__(self, matrix, weights, types):
        """
        Score alternatives from decision matrix `matrix` using criteria weights `weights` and 
        criteria types `types`

        Parameters
        ----------
            matrix : ndarray
                decision matrix with performance values for m alternatives in rows and n criteria 
                in columns
            weights : ndarray
                matrix with criteria weights vectors with number of columns equal to 
                number of columns n of `matrix`
            types : ndarray
                vector with criteria types containing values of 1 for profit criteria and -1 for 
                cost criteria with size equal to number of columns n of `matrix`
        """
        pass

    @staticmethod
    def _verify_input_data(matrix, weights, types):
        m, n = matrix.shape
        # if weights are vector with one dimension
        if len(weights.shape) == 1:
            if len(weights) != n:
                raise ValueError('The size of the weight vector must be the same as the number of criteria')
        # if weights are two-dimensional matrix containing many weight vectors in rows
        elif len(weights.shape) == 2:
            if weights.shape[1] != n:
                raise ValueError('The number of columns of matrix with weight vectors must be the same as the number of criteria')
        if len(types) != n:
            raise ValueError('The size of the types vector must be the same as the number of criteria')
        check_types = np.all((types == 1) | (types == -1))
        if check_types == False:
            raise ValueError('Criteria types can only have a value of 1 for profits and -1 for costs')


import itertools
import numpy as np
from .mcda_method import MCDA_method


class PROMETHEE_II(MCDA_method):
    def __init__(self):
        """
        Create the PROMETHEE II method object
        """
        pass


    def __call__(self, matrix, weights, types, preference_functions = None, p = None, q = None):
        PROMETHEE_II._verify_input_data(matrix, weights, types)

        if preference_functions is None:
            preference_functions = [self._usual_function for pf in range(len(weights))]
            
        if p is None:
            u = np.sqrt(np.sum(np.square(np.mean(matrix, axis = 0) - matrix), axis = 0) / matrix.shape[0])
            p = 2 * u
            
        if q is None:
            u = np.sqrt(np.sum(np.square(np.mean(matrix, axis = 0) - matrix), axis = 0) / matrix.shape[0])
            q = 0.5 * u
            
        if len(preference_functions) != np.shape(matrix)[1]:
            raise ValueError('The list of preference functions must be equal in length to the number of criteria')
        if len(p) != np.shape(matrix)[1]:
            raise ValueError('The length of the vector p must be equal to the number of criteria')
        if len(q) != np.shape(matrix)[1]:
            raise ValueError('The length of the vector q must be equal to the number of criteria')

        return PROMETHEE_II._promethee_II(self, matrix, weights, types, preference_functions, p, q)


    # Preference function type 1 (Usual criterion) requires no parameters
    # alternatives are indifferent only if they are equal to each other
    # otherwise there is a strong preference for one of them
    def _usual_function(self, d, p, q):
        if d <= 0:
            return 0
        else:
            return 1

    # Preference function type 2 (U-shape criterion) requires indifference threshold (q)
    def _ushape_function(self, d, p, q):
        if d <= q:
            return 0
        else:
            return 1

    # Preference function type 3 (V-shape criterion) requires threshold of absolute preference (p) 
    def _vshape_function(self, d, p, q):
        if d <= 0:
            return 0
        elif 0 <= d <= p:
            return d / p
        elif d > p:
            return 1

    # preference function type 4 (Level criterion) requires both preference and indifference thresholds (p and q)
    def _level_function(self, d, p, q):
        if d <= q:
            return 0
        elif q < d <= p:
            return 0.5
        elif d > p:
            return 1

    # Preference function type 5 (V-shape with indifference criterion also known as linear)
    # requires both preference and indifference thresholds (p and q)
    def _linear_function(self, d, p, q):
        if d <= q:
            return 0
        elif q < d <= p:
            return (d - q) / (p - q)
        elif d > p:
            return 1

    # preference function type 6 (Gaussian criterion)
    # requires to fix parameter s which is an intermediate value between q and p
    def _gaussian_function(self, d, p, q):
        if d <= 0:
            return 0
        elif d > 0:
            s = (p + q) / 2
            return 1 - np.exp(-((d ** 2) / (2 * s ** 2)))


    @staticmethod
    def _promethee_II(self, matrix, weights, types, preference_functions, p, q):
        """
        Score alternatives provided in the decision matrix `matrix` using criteria `weights` and criteria `types`.
        
        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Criteria weights. The sum of weights must be equal to 1.
            types: ndarray
                Criteria types. Profit criteria are represented by 1 and cost by -1.
            preference_functions : list
                List with methods containing preference functions for calculating the
                preference degree for each criterion.
            p : ndarray
                Vector with values representing the threshold of absolute preference.
            q : ndarray
                Vector with values representing the threshold of indifference.
        
        Returns
        --------
            ndrarray
                Preference values of each alternative. The best alternative has the highest preference value. 
        
        Examples
        ----------
        >>> promethee_II = PROMETHEE_II()
        >>> preference_functions = [promethee_II._linear_function for pf in range(len(weights))]
        >>> u = np.sqrt(np.sum(np.square(np.mean(matrix, axis = 0) - matrix), axis = 0) / matrix.shape[0])
        >>> p = 2 * u
        >>> q = 0.5 * u
        >>> pref = promethee_II(matrix, weights, types, preference_functions, p = p, q = q)
        >>> rank = rank_preferences(pref, reverse = True)
        """

        m, n = matrix.shape

        H = np.zeros((m, m))

        # A preference index of two options, i-th and k-th, is calculated
        for i, k, j in itertools.product(range(m), range(m), range(n)):
            H[i, k] += preference_functions[j](types[j] * (matrix[i, j] - matrix[k, j]), p[j], q[j]) * weights[j]

        # Output phi_plus and input phi_minus dominance flows are determined
        phi_plus = np.sum(H, axis = 1) / (m - 1)
        phi_min = np.sum(H, axis = 0) / (m - 1)

        # The net dominance flow phi_net is calculated
        phi_net = phi_plus - phi_min
        return phi_net


import unittest
import numpy as np
from pyrepo_mcda import weighting_methods as mcda_weights
from pyrepo_mcda import normalizations as norms



# Test for CRITIC weighting
class Test_CRITIC(unittest.TestCase):

    def test_critic(self):
        """Test based on paper Tuş, A., & Aytaç Adalı, E. (2019). The new combination with CRITIC and WASPAS methods 
        for the time and attendance software selection problem. Opsearch, 56(2), 528-538. 
        DOI: https://doi.org/10.1007/s12597-019-00371-6"""

        matrix = np.array([[5000, 3, 3, 4, 3, 2],
        [680, 5, 3, 2, 2, 1],
        [2000, 3, 2, 3, 4, 3],
        [600, 4, 3, 1, 2, 2],
        [800, 2, 4, 3, 3, 4]])

        types = np.array([-1, 1, 1, 1, 1, 1])

        test_result = mcda_weights.critic_weighting(matrix)
        real_result = np.array([0.157, 0.249, 0.168, 0.121, 0.154, 0.151])
        self.assertEqual(list(np.round(test_result, 3)), list(real_result))


# Test for MEREC weighting
class Test_MEREC(unittest.TestCase):

    def test_merec(self):
        """Test based on paper Keshavarz-Ghorabaee, M., Amiri, M., Zavadskas, E. K., Turskis, Z., & Antucheviciene, 
        J. (2021). Determination of objective weights using a new method based on the removal 
        effects of criteria (MEREC). Symmetry, 13(4), 525. DOI: https://doi.org/10.3390/sym13040525"""

        matrix = np.array([[450, 8000, 54, 145],
        [10, 9100, 2, 160],
        [100, 8200, 31, 153],
        [220, 9300, 1, 162],
        [5, 8400, 23, 158]])

        types = np.array([1, 1, -1, -1])

        test_result = mcda_weights.merec_weighting(matrix, types)
        real_result = np.array([0.5752, 0.0141, 0.4016, 0.0091])
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))


# Test for Entropy weighting
class Test_Entropy(unittest.TestCase):

    def test_Entropy(self):
        """Test based on paper Xu, X. (2004). A note on the subjective and objective integrated approach to 
        determine attribute weights. European Journal of Operational Research, 156(2), 
        530-532. DOI: https://doi.org/10.1016/S0377-2217(03)00146-2"""

        matrix = np.array([[30, 30, 38, 29],
        [19, 54, 86, 29],
        [19, 15, 85, 28.9],
        [68, 70, 60, 29]])

        types = np.array([1, 1, 1, 1])

        test_result = mcda_weights.entropy_weighting(matrix)
        real_result = np.array([0.4630, 0.3992, 0.1378, 0.0000])
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))

    def test_Entropy2(self):
        """Test based on paper Zavadskas, E. K., & Podvezko, V. (2016). Integrated determination of objective 
        criteria weights in MCDM. International Journal of Information Technology & Decision 
        Making, 15(02), 267-283. DOI: https://EconPapers.repec.org/RePEc:wsi:ijitdm:v:15:y:2016:i:02:n:s0219622016500036"""

        matrix = np.array([[3.0, 100, 10, 7],
        [2.5, 80, 8, 5],
        [1.8, 50, 20, 11],
        [2.2, 70, 12, 9]])

        types = np.array([-1, 1, -1, 1])

        test_result = mcda_weights.entropy_weighting(matrix)
        real_result = np.array([0.1146, 0.1981, 0.4185, 0.2689])
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))

    def test_Entropy3(self):
        """Test based on paper Ersoy, Y. (2021). Equipment selection for an e-commerce company using entropy-based 
        topsis, edas and codas methods during the COVID-19. LogForum, 17(3). DOI: http://doi.org/10.17270/J.LOG.2021.603"""

        matrix = np.array([[256, 8, 41, 1.6, 1.77, 7347.16],
        [256, 8, 32, 1.0, 1.8, 6919.99],
        [256, 8, 53, 1.6, 1.9, 8400],
        [256, 8, 41, 1.0, 1.75, 6808.9],
        [512, 8, 35, 1.6, 1.7, 8479.99],
        [256, 4, 35, 1.6, 1.7, 7499.99]])

        types = np.array([-1, 1, -1, 1])
        test_result = mcda_weights.entropy_weighting(matrix)

        real_result = np.array([0.405, 0.221, 0.134, 0.199, 0.007, 0.034])
        self.assertEqual(list(np.round(test_result, 3)), list(real_result))

    def test_Entropy4(self):
        """Test based on paper Lee, H. C., & Chang, C. T. (2018). Comparative analysis of MCDM 
        methods for ranking renewable energy sources in Taiwan. Renewable and Sustainable Energy 
        Reviews, 92, 883-896. DOI: https://doi.org/10.1016/j.rser.2018.05.007"""

        matrix = np.array([[4550, 30, 6.74, 20, 15, 5, 85, 150, 0.87, 4.76],
        [3005, 60.86, 2.4, 35, 27, 4, 26, 200, 0.17, 4.51],
        [2040, 14.85, 1.7, 90, 25, 5, 26, 500, 0.27, 4.19],
        [3370, 99.4, 3.25, 25.3, 54, 3, 45, 222, 0.21, 3.78],
        [3920, 112.6, 4.93, 11.4, 71.7, 2, 50, 100, 0.25, 4.11]])

        types = np.array([-1, -1, -1, 1, 1, 1, -1, -1, 1, 1])

        test_result = mcda_weights.entropy_weighting(matrix)
        real_result = np.array([0.026, 0.154, 0.089, 0.199, 0.115, 0.04, 0.08, 0.123, 0.172, 0.002])
        self.assertEqual(list(np.round(test_result, 3)), list(real_result))
        

# Test for CILOS weighting
class Test_CILOS(unittest.TestCase):

    def test_cilos(self):
        """Test based on paper Alinezhad, A., & Khalili, J. (2019). New methods and applications in multiple 
        attribute decision making (MADM) (Vol. 277). Cham: Springer. DOI: https://doi.org/10.1007/978-3-030-15009-9"""

        matrix = np.array([[3, 100, 10, 7],
        [2.500, 80, 8, 5],
        [1.800, 50, 20, 11],
        [2.200, 70, 12, 9]])

        types = np.array([-1, 1, -1, 1])

        test_result = mcda_weights.cilos_weighting(matrix, types)
        
        real_result = np.array([0.334, 0.220, 0.196, 0.250])
        self.assertEqual(list(np.round(test_result, 3)), list(real_result))


    def test_cilos2(self):
        """Test based on paper Zavadskas, E. K., & Podvezko, V. (2016). Integrated determination of objective 
        criteria weights in MCDM. International Journal of Information Technology & Decision 
        Making, 15(02), 267-283. DOI: https://EconPapers.repec.org/RePEc:wsi:ijitdm:v:15:y:2016:i:02:n:s0219622016500036"""

        matrix = np.array([[0.6, 100, 0.8, 7],
        [0.72, 80, 1, 5],
        [1, 50, 0.4, 11],
        [0.818, 70, 0.667, 9]])

        types = np.array([1, 1, 1, 1])

        test_result = mcda_weights.cilos_weighting(matrix, types)
        real_result = np.array([0.3343, 0.2199, 0.1957, 0.2501])
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))


# Test for IDOCRIW weighting
class Test_IDOCRIW(unittest.TestCase):

    def test_idocriw(self):
        """Test based on paper Zavadskas, E. K., & Podvezko, V. (2016). Integrated determination of objective 
        criteria weights in MCDM. International Journal of Information Technology & Decision 
        Making, 15(02), 267-283. DOI: https://EconPapers.repec.org/RePEc:wsi:ijitdm:v:15:y:2016:i:02:n:s0219622016500036"""

        matrix = np.array([[3.0, 100, 10, 7],
        [2.5, 80, 8, 5],
        [1.8, 50, 20, 11],
        [2.2, 70, 12, 9]])

        types = np.array([-1, 1, -1, 1])

        test_result = mcda_weights.idocriw_weighting(matrix, types)
        real_result = np.array([0.1658, 0.1886, 0.35455, 0.2911])
        self.assertEqual(list(np.round(test_result, 3)), list(np.round(real_result, 3)))


# Test for Angle weighting
class Test_Angle(unittest.TestCase):

    def test_angle(self):
        """Test based on paper Shuai, D., Zongzhun, Z., Yongji, W., & Lei, L. (2012, May). A new angular method to 
        determine the objective weights. In 2012 24th Chinese Control and Decision Conference 
        (CCDC) (pp. 3889-3892). IEEE. DOI: https://doi.org/10.1109/CCDC.2012.6244621"""

        matrix = np.array([[30, 30, 38, 29],
        [19, 54, 86, 29],
        [19, 15, 85, 28.9],
        [68, 70, 60, 29]])

        types = np.array([1, 1, 1, 1])

        test_result = mcda_weights.angle_weighting(matrix, types)
        real_result = np.array([0.4150, 0.3612, 0.2227, 0.0012])
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))


# Test for Coefficient of Variation weighting
class Test_Coeff_var(unittest.TestCase):

    def test_coeff_var(self):
        """Test based on paper Shuai, D., Zongzhun, Z., Yongji, W., & Lei, L. (2012, May). A new angular method to 
        determine the objective weights. In 2012 24th Chinese Control and Decision Conference 
        (CCDC) (pp. 3889-3892). IEEE. DOI: https://doi.org/10.1109/CCDC.2012.6244621"""

        matrix = np.array([[30, 30, 38, 29],
        [19, 54, 86, 29],
        [19, 15, 85, 28.9],
        [68, 70, 60, 29]])
        
        types = np.array([1, 1, 1, 1])

        test_result = mcda_weights.coeff_var_weighting(matrix)
        real_result = np.array([0.4258, 0.3610, 0.2121, 0.0011])
        self.assertEqual(list(np.round(test_result, 4)), list(real_result))


# Test for Standard Deviation weighting
class Test_STD(unittest.TestCase):

    def test_std(self):
        """Test based on paper Sałabun, W., Wątróbski, J., & Shekhovtsov, A. (2020). Are mcda methods benchmarkable? 
        a comparative study of topsis, vikor, copras, and promethee ii methods. Symmetry, 12(9), 
        1549. DOI: https://doi.org/10.3390/sym12091549"""

        matrix = np.array([[0.619, 0.449, 0.447],
        [0.862, 0.466, 0.006],
        [0.458, 0.698, 0.771],
        [0.777, 0.631, 0.491],
        [0.567, 0.992, 0.968]])
        
        types = np.array([1, 1, 1])

        test_result = mcda_weights.std_weighting(matrix)
        real_result = np.array([0.217, 0.294, 0.488])
        self.assertEqual(list(np.round(test_result, 3)), list(real_result))


# Test for Equal weighting
class Test_equal(unittest.TestCase):

    def test_equal(self):
        """Test based on paper Sałabun, W., Wątróbski, J., & Shekhovtsov, A. (2020). Are mcda methods benchmarkable? 
        a comparative study of topsis, vikor, copras, and promethee ii methods. Symmetry, 12(9), 
        1549. DOI: https://doi.org/10.3390/sym12091549"""

        matrix = np.array([[0.619, 0.449, 0.447],
        [0.862, 0.466, 0.006],
        [0.458, 0.698, 0.771],
        [0.777, 0.631, 0.491],
        [0.567, 0.992, 0.968]])
        
        types = np.array([1, 1, 1])

        test_result = mcda_weights.equal_weighting(matrix)
        real_result = np.array([0.333, 0.333, 0.333])
        self.assertEqual(list(np.round(test_result, 3)), list(real_result))


# Test for Statistical variance weighting
class Test_stat_var(unittest.TestCase):

    def test_stat_var(self):
        """Test based on paper Sałabun, W., Wątróbski, J., & Shekhovtsov, A. (2020). Are mcda methods benchmarkable? 
        a comparative study of topsis, vikor, copras, and promethee ii methods. Symmetry, 12(9), 
        1549. DOI: https://doi.org/10.3390/sym12091549"""

        matrix = np.array([[0.619, 0.449, 0.447],
        [0.862, 0.466, 0.006],
        [0.458, 0.698, 0.771],
        [0.777, 0.631, 0.491],
        [0.567, 0.992, 0.968]])
        
        types = np.array([1, 1, 1])

        test_result = mcda_weights.stat_var_weighting(matrix)
        xn = norms.minmax_normalization(matrix, np.ones(matrix.shape[1]))
        real_result = np.var(xn, axis = 0) / np.sum(np.var(xn, axis = 0))
        self.assertEqual(list(np.round(test_result, 4)), list(np.round(real_result, 4)))


# Test for Gini coefficient-based weighting
class Test_gini(unittest.TestCase):

    def test_gini(self):
        """Test based on paper Bączkiewicz, A., Wątróbski, J., Kizielewicz, B., & Sałabun, W. (2021). Towards 
        Reliable Results-A Comparative Analysis of Selected MCDA Techniques in the Camera 
        Selection Problem. In Information Technology for Management: Business and Social Issues 
        (pp. 143-165). Springer, Cham. DOI: 10.1007/978-3-030-98997-2_7"""

        matrix = np.array([[29.4, 83, 47, 114, 12, 30, 120, 240, 170, 90, 1717.75],
        [30, 38.1, 124.7, 117, 16, 60, 60, 60, 93, 70, 2389],
        [29.28, 59.27, 41.13, 58, 16, 30, 60, 120, 170, 78, 239.99],
        [33.6, 71, 55, 159, 23.6, 60, 240, 240, 132, 140, 2099],
        [21, 59, 41, 66, 16, 24, 60, 120, 170, 70, 439],
        [35, 65, 42, 134, 12, 60, 240, 240, 145, 60, 1087],
        [47, 79, 54, 158, 19, 60, 120, 120, 360, 72, 2499],
        [28.3, 62.3, 44.9, 116, 12, 30, 60, 60, 130, 90, 999.99],
        [36.9, 28.6, 121.6, 130, 12, 60, 120, 120, 80, 80, 1099],
        [32, 59, 41, 60, 16, 30, 120, 120, 170, 60, 302.96],
        [28.4, 66.3, 48.6, 126, 12, 60, 240, 240, 132, 135, 1629],
        [29.8, 46, 113, 47, 18, 50, 50, 50, 360, 72, 2099],
        [20.2, 64, 80, 70, 8, 24, 60, 120, 166, 480, 699.99],
        [33, 60, 44, 59, 12, 30, 60, 120, 170, 90, 388],
        [29, 59, 41, 55, 16, 30, 60, 120, 170, 120, 299],
        [29, 59, 41, 182, 12, 30, 30, 60, 94, 140, 249],
        [29.8, 59.2, 41, 65, 16, 30, 60, 120, 160, 90, 219.99],
        [28.8, 62.5, 41, 70, 12, 60, 120, 120, 170, 138, 1399.99],
        [24, 40, 59, 60, 12, 10, 30, 30, 140, 78, 269.99],
        [30, 60, 45, 201, 16, 30, 30, 30, 170, 90, 199.99]])
        
        types = np.array([-1, -1, -1, -1, 1, 1, 1, 1, 1, 1, -1])

        test_result = mcda_weights.gini_weighting(matrix)
        
        real_result = np.array([0.0362, 0.0437, 0.0848, 0.0984, 0.0480, 0.0842, 0.1379, 0.1125, 0.0745, 0.1107, 0.1690])
        self.assertEqual(list(np.round(test_result, 4)), list(np.round(real_result, 4)))
        
        
        
def main():
    # Test of the CRITIC weighting method
    test_critic = Test_CRITIC()
    test_critic.test_critic()

    # Test of the MEREC weighting method
    test_merec = Test_MEREC()
    test_merec.test_merec()

    # Test of the Entropy weighting method
    test_entropy = Test_Entropy()
    test_entropy.test_Entropy()
    test_entropy.test_Entropy2()
    test_entropy.test_Entropy3()
    test_entropy.test_Entropy4()

    # Test of the CILOS weighting method
    test_cilos = Test_CILOS()
    test_cilos.test_cilos()
    test_cilos.test_cilos2()

    # Test of the IDOCRIW weighting method
    test_idocriw = Test_IDOCRIW()
    test_idocriw.test_idocriw()

    # Test of the Angle weighting method
    test_angle = Test_Angle()
    test_angle.test_angle()

    # Test of the Coefficient of variation weighting method
    test_coeff_var = Test_Coeff_var()
    test_coeff_var.test_coeff_var()

    # Test of the Standard deviation weighting method
    test_std = Test_STD()
    test_std.test_std()

    # Test of the Equal weighting method
    test_equal = Test_equal()
    test_equal.test_equal()

    # Test of the Statistical variance weighting method
    test_stat_var = Test_stat_var()
    test_stat_var.test_stat_var()
    
    # Test of the Gini coefficient-based weighting method
    test_gini = Test_gini()
    test_gini.test_gini()

if __name__ == '__main__':
    main()




import numpy as np

from ..normalizations import minmax_normalization
from ..distance_metrics import euclidean
from .mcda_method import MCDA_method

class TOPSIS(MCDA_method):
    def __init__(self, normalization_method = minmax_normalization, distance_metric = euclidean):
        """
        Create the TOPSIS method object and select normalization method `normalization_method` and
        distance metric `distance metric`.

        Parameters
        -----------
            normalization_method : function
                method for decision matrix normalization chosen from `normalizations`

            distance_metric : functions
                method for calculating the distance between two vectors
        """
        self.normalization_method = normalization_method
        self.distance_metric = distance_metric


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` with m alternatives in rows and 
        n criteria in columns using criteria `weights` and criteria `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ---------
        >>> topsis = TOPSIS(normalization_method = minmax_normalization, distance_metric = euclidean)
        >>> pref = topsis(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        TOPSIS._verify_input_data(matrix, weights, types)
        return TOPSIS._topsis(matrix, weights, types, self.normalization_method, self.distance_metric)


    @staticmethod
    def _topsis(matrix, weights, types, normalization_method, distance_metric):
        # Normalize matrix using chosen normalization (for example linear normalization)
        norm_matrix = normalization_method(matrix, types)

        # Multiply all rows of normalized matrix by weights
        weighted_matrix = norm_matrix * weights

        # Calculate vectors of PIS (ideal solution) and NIS (anti-ideal solution)
        pis = np.max(weighted_matrix, axis=0)
        nis = np.min(weighted_matrix, axis=0)

        # Calculate chosen distance of every alternative from PIS and NIS using chosen distance metric `distance_metric` from `distance_metrics`
        Dp = np.array([distance_metric(x, pis) for x in weighted_matrix])
        Dm = np.array([distance_metric(x, nis) for x in weighted_matrix])

        C = Dm / (Dm + Dp)
        return C


import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd


# bar (column) chart
def plot_barplot(df_plot, legend_title):
    """
    Visualization method to display column chart of alternatives rankings obtained with 
    different methods.

    Parameters
    ----------
        df_plot : DataFrame
            DataFrame containing rankings of alternatives obtained with different methods.
            The particular rankings are included in subsequent columns of DataFrame.

        title : str
            Title of the legend (Name of group of explored methods, for example MCDA methods or Distance metrics).

    Examples
    ----------
    >>> plot_barplot(df_plot, legend_title='MCDA methods')
    """
    step = 1
    list_rank = np.arange(1, len(df_plot) + 1, step)

    ax = df_plot.plot(kind='bar', width = 0.8, stacked=False, edgecolor = 'black', figsize = (9,4))
    ax.set_xlabel('Alternatives', fontsize = 12)
    ax.set_ylabel('Rank', fontsize = 12)
    ax.set_yticks(list_rank)

    ax.set_xticklabels(df_plot.index, rotation = 'horizontal')
    ax.tick_params(axis = 'both', labelsize = 12)
    y_ticks = ax.yaxis.get_major_ticks()
    ax.set_ylim(0, len(df_plot) + 1)

    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc='lower left',
    ncol=4, mode="expand", borderaxespad=0., edgecolor = 'black', title = legend_title, fontsize = 12)

    ax.grid(True, linestyle = ':')
    ax.set_axisbelow(True)
    plt.tight_layout()
    legend_title = legend_title.replace("$", "")
    legend_title = legend_title.replace("{", "")
    legend_title = legend_title.replace("}", "")
    plt.savefig('./results/' + 'bar_chart_' + legend_title + '.eps')
    plt.show()


# bar (column) chart for sensitivity analysis
def plot_barplot_sensitivity(df_plot, method_name, criterion_name, filename = ""):
    """
    Visualization method to display column chart of alternatives rankings obtained with 
    modification of weight of given criterion.

    Parameters
    ----------
        df_plot : DataFrame
            DataFrame containing rankings of alternatives obtained with different weights of
            selected criterion. The particular rankings are contained in subsequent columns of 
            DataFrame.

        method_name : str
            Name of chosen MCDA method, i.e. `TOPSIS`, `VIKOR`, `CODAS`, `WASPAS`, `MULTIMOORA`, `MABAC`, `EDAS`, `SPOTIS`
        
        criterion_name : str
            Name of chosen criterion whose weight is modified

        filename : str
            Name of file to save this chart

    Examples
    -----------
    >>> plot_barplot_sensitivity(df_plot, method_name, criterion_name, filename)
    """
    step = 1
    list_rank = np.arange(1, len(df_plot) + 1, step)

    ax = df_plot.plot(kind='bar', width = 0.8, stacked=False, edgecolor = 'black', figsize = (9,4))
    ax.set_xlabel('Alternatives', fontsize = 12)
    ax.set_ylabel('Rank', fontsize = 12)
    ax.set_yticks(list_rank)

    ax.set_xticklabels(df_plot.index, rotation = 'horizontal')
    ax.tick_params(axis='both', labelsize=12)
    y_ticks = ax.yaxis.get_major_ticks()
    ax.set_ylim(0, len(df_plot) + 1)
    ax.set_title(method_name + ', modification of ' + criterion_name + ' weights')

    plt.legend(bbox_to_anchor=(1.0, 0.82, 0.3, 0.2), loc='upper left', title = 'Weights change', edgecolor = 'black', fontsize = 12)

    ax.grid(True, linestyle = ':')
    ax.set_axisbelow(True)
    plt.tight_layout()
    criterion_name = criterion_name.replace("$", "")
    criterion_name = criterion_name.replace("{", "")
    criterion_name = criterion_name.replace("}", "")
    plt.savefig('./results/' + 'sensitivity_' + 'hist_' + method_name + '_' + criterion_name + '_' + filename + '.eps')
    plt.show()


# plot line chart for sensitivity analysis
def plot_lineplot_sensitivity(data_sens, method_name, criterion_name, x_title, filename = ""):
    """
    Visualization method to display line chart of alternatives rankings obtained with 
    modification of weight of given criterion.

    Parameters
    ----------
        df_plot : DataFrame
            DataFrame containing rankings of alternatives obtained with different weight of
            selected criterion. The particular rankings are contained in subsequent columns of 
            DataFrame.

        method_name : str
            Name of chosen MCDA method, i.e. `TOPSIS`, `VIKOR`, `CODAS`, `WASPAS`, `MULTIMOORA`, `MABAC`, `EDAS`, `SPOTIS`
        
        criterion_name : str
            Name of chosen criterion whose weight is modified
        
        x_title : str
            Title of x axis

        filename : str
            Name of file to save this chart

    Examples
    ----------
    >>> plot_lineplot_sensitivity(df_plot, method_name, criterion_name, x_title, filename)
    """
    plt.figure(figsize = (6, 3))
    for j in range(data_sens.shape[0]):
        
        plt.plot(data_sens.iloc[j, :], linewidth = 2)
        ax = plt.gca()
        y_min, y_max = ax.get_ylim()
        x_min, x_max = ax.get_xlim()
        plt.annotate(data_sens.index[j], (x_max, data_sens.iloc[j, -1]),
                        fontsize = 12, style='italic',
                        horizontalalignment='left')

    plt.xlabel(x_title, fontsize = 12)
    plt.ylabel("Rank", fontsize = 12)
    plt.yticks(fontsize = 12)
    plt.xticks(fontsize = 12)
    plt.title(method_name + ', modification of ' + criterion_name + ' weight')
    plt.grid(True, linestyle = ':')
    plt.tight_layout()
    criterion_name = criterion_name.replace("$", "")
    criterion_name = criterion_name.replace("{", "")
    criterion_name = criterion_name.replace("}", "")
    plt.savefig('./results/' + 'sensitivity_' + 'lineplot_' + method_name + '_' + criterion_name + '_' + filename + '.eps')
    plt.show()


# heat maps with correlations
def draw_heatmap(df_new_heatmap, title):
    """
    Visualization method to display heatmap with correlations of compared rankings generated using different methods
    
    Parameters
    ----------
        data : DataFrame
            DataFrame with correlation values between compared rankings

        title : str
            title of chart containing name of used correlation coefficient

    Examples
    ---------
    >>> draw_heatmap(df_new_heatmap, title)
    """
    plt.figure(figsize = (8, 5))
    sns.set(font_scale = 1.2)
    heatmap = sns.heatmap(df_new_heatmap, annot=True, fmt=".2f", cmap="PuBu",
                          linewidth=0.5, linecolor='w')
    plt.yticks(va="center")
    plt.xlabel('Methods')
    plt.title('Correlation: ' + title)
    plt.tight_layout()
    title = title.replace("$", "")
    plt.savefig('./results/' + 'correlations_' + title + '.eps')
    plt.show()


# radar chart
def plot_radar(data, title, j):
    """
    Visualization method to display rankings of alternatives obtained with different methods
    on the radar chart.

    Parameters
    -----------
        data : DataFrame
            DataFrame containing containing rankings of alternatives obtained with different 
            methods. The particular rankings are contained in subsequent columns of DataFrame.

        title : str
            Chart title

        j : int
            Index of criterion chosen for weight modification in sensitivity analysis

    Examples
    ----------
    >>> plot_radar(data, title, j)
    """
    fig=plt.figure()
    ax = fig.add_subplot(111, polar = True)

    for col in list(data.columns):
        labels=np.array(list(data.index))
        stats = data.loc[labels, col].values

        angles=np.linspace(0, 2*np.pi, len(labels), endpoint=False)
        # close the plot
        stats=np.concatenate((stats,[stats[0]]))
        angles=np.concatenate((angles,[angles[0]]))
    
        lista = list(data.index)
        lista.append(data.index[0])
        labels=np.array(lista)

        ax.plot(angles, stats, '-o', linewidth=2)
    
    ax.set_thetagrids(angles * 180/np.pi, labels)
    ax.grid(True)
    ax.set_axisbelow(True)
    plt.legend(data.columns, bbox_to_anchor=(1.0, 0.95, 0.4, 0.2), loc='upper left')
    plt.title(title)
    plt.tight_layout()
    plt.savefig('./results/' + 'radar_chart_C' + str(j + 1) + '.eps')
    plt.show()

# Examplary visualization method
def plot_boxplot(data, title):
    """
    Display boxplot showing distribution of preference values determined with different methods.

    Parameters
    ----------
        data : dataframe
            dataframe with correlation values between compared rankings

        title : str
            Title of chart.

    Examples
    ----------
    >>> plot_boxplot(data, title)
    """
    
    df_melted = pd.melt(data)
    plt.figure(figsize = (7, 4))
    ax = sns.boxplot(x = 'variable', y = 'value', data = df_melted, width = 0.4)
    ax.grid(True, linestyle = '--')
    ax.set_axisbelow(True)
    ax.set_xlabel('Alternatives', fontsize = 12)
    ax.set_ylabel('Preference distribution', fontsize = 12)
    plt.title(title)
    plt.tight_layout()
    plt.savefig('./results/' + 'TOPSIS_boxplot' + '.eps')
    plt.show()


# plot box chart of results obtained in robustness analysis showing intervals of alternatives performance values for particular ranks
def plot_boxplot_simulation(data, x, y, hue, xtitle, ytitle, title, filename):
    """
    Visualization method to display box chart of results obtained in robustness analysis. It shows
    intervals of alternatives performance values for particular positions in ranking.

    Parameters
    ----------
        data : DataFrame
            DataFrame containing results

        x : str
            Name of column in DataFrame with variable names in axis x on chart

        y : str
            Name of column in DataFrame with variable values in axis y on chart
        hue : str
            Name of hue, that determines how the data are plotted

        xtitle : str
            Name of axis x title

        ytitle : str
            Name of axis y title

        title : str
            Chart title

        filename : str
            Name of file in which chart will be saved

    Examples
    ----------
    >>> plot_boxplot_simulation(data, x, y, hue , xtitle, ytitle, title, filename)

    """
    plt.figure(figsize = (9,5))
    ax = sns.boxplot(x = x, y = y, hue = hue, palette='husl', data = data)
    ax.set_xlabel(xtitle, fontsize = 12)
    ax.set_ylabel(ytitle, fontsize = 12)
    ax.grid(True, linestyle = ':')
    ax.set_axisbelow(True)
    plt.legend(bbox_to_anchor=(1.0, 0.82, 0.3, 0.2), loc='upper left', title = 'Rank', edgecolor = 'black', fontsize = 12)
    plt.title(title)
    plt.tight_layout()
    plt.savefig('./results/' + filename + '.eps')
    plt.show()


from . import mcda_methods
from . import additions
from . import compromise_rankings
from . import correlations
from . import distance_metrics
from . import normalizations
from . import sensitivity_analysis_weights_percentages
from . import sensitivity_analysis_weights_values
from . import weighting_methods


from . import mcda_method
from .ahp import AHP
from .aras import ARAS
from .cocoso import COCOSO
from .codas import CODAS
from .copras import COPRAS
from .cradis import CRADIS
from .edas import EDAS
from .mabac import MABAC
from .marcos import MARCOS
from .multimoora import MULTIMOORA
from .multimoora import MULTIMOORA_RS
from .promethee import PROMETHEE_II
from .prosa_c import PROSA_C
from .saw import SAW
from .spotis import SPOTIS
from .topsis import TOPSIS
from .vikor import VIKOR
from .waspas import WASPAS
from .vikor_smaa import VIKOR_SMAA


import numpy as np
from ..normalizations import minmax_normalization
from .mcda_method import MCDA_method


class MABAC(MCDA_method):
    def __init__(self, normalization_method = minmax_normalization):
        """
        Create the MABAC method object and select normalization method `normalization_method` from
        `normalizations`. The default normalization method for MABAC method is `minmax_normalization`

        Parameters
        -----------
            normalization_method : function
                method for decision matrix normalization chosen from `normalizations`
        """
        self.normalization_method = normalization_method


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` using criteria `weights` and criteria `types`.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        --------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ---------
        >>> mabac = MABAC(normalization_method = minmax_normalization)
        >>> pref = mabac(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """
        MABAC._verify_input_data(matrix, weights, types)
        return MABAC._mabac(matrix, weights, types, self.normalization_method)


    def _mabac(matrix, weights, types, normalization_method):
        m, n = matrix.shape
        # Normalize the decision matrix
        norm_matrix = normalization_method(matrix, types)
        # Calculate elements from the weighted matrix
        V = weights * (norm_matrix + 1)
        # Determine the border approximation area matrix
        G = np.product(V, axis = 0) ** (1/m)
        # Calculate distance of alternatives from the border approximation area for the matrix elements
        Q = V - G
        # Calculate the sum of distance of alternatives from the border approximation areas
        # By calculating the sum of elements from Q matrix by rows we obtain the final values of
        # the criterion functions of the alternatives
        S = np.sum(Q, axis = 1)
        return S




import numpy as np

from ..normalizations import minmax_normalization
from .mcda_method import MCDA_method


class COCOSO(MCDA_method):
    def __init__(self, normalization_method = minmax_normalization, lambda_param = 0.5):
        """
        Create the COCOSO method object and select value of lambda parameter called `lambda_param`.
        By default, the lambda parameter is equal to 0.5

        Parameters
        -----------
            lambda_param : parameter chosen by decision makers, usually is equal to 0.5. It determines
            flexibility and stability of the proposed CoCoSo
        """

        self.normalization_method = normalization_method
        self.lambda_param = lambda_param


    def __call__(self, matrix, weights, types):
        """
        Score alternatives provided in decision matrix `matrix` with m alternatives in rows and 
        n criteria in columns using criteria `weights` and criteria `types`.

        Parameters
        ----------
            matrix : ndarray
                Decision matrix with m alternatives in rows and n criteria in columns.
            weights: ndarray
                Vector with criteria weights. Sum of weights must be equal to 1.
            types: ndarray
                Vector with criteria types. Profit criteria are represented by 1 and cost by -1.

        Returns
        -------
            ndrarray
                Vector with preference values of each alternative. The best alternative has the highest preference value. 

        Examples
        ---------
        >>> cocoso = COCOSO(lambda_param = lambda_param)
        >>> pref = cocoso(matrix, weights, types)
        >>> rank = rank_preferences(pref, reverse = True)
        """

        COCOSO._verify_input_data(matrix, weights, types)
        return COCOSO._cocoso(matrix, weights, types, self.normalization_method, self.lambda_param)


    @staticmethod
    def _cocoso(matrix, weights, types, normalization_method = minmax_normalization, lambda_param = 0.5):
        # Normalize matrix using chosen normalization. minmax_normalization is default
        norm_matrix = normalization_method(matrix, types)
        S = np.sum(weights * norm_matrix, axis = 1)
        P = np.sum(norm_matrix**weights, axis = 1)

        kia = (P + S) / np.sum(P + S)
        kib = S / np.min(S) + P / np.min(P)
        kic = (lambda_param * S + (1 - lambda_param) * P) / (lambda_param * np.max(S) + (1 - lambda_param) * np.max(P))
        K = (kia * kib * kic)**(1/3) + 1/3*(kia + kib + kic)
        return K


import copy
import numpy as np
import pandas as pd

from .additions import rank_preferences


class Sensitivity_analysis_weights_percentages():
    def __init__(self):
        """
        Create object of `Sensitivity_analysis_weights_percentages` method.
        """
        pass


    def __call__(self, matrix, weights, types, percentages, method, j, dir_list):
        """
        Method for sensitivity analysis. This method determines rankings of alternatives using chosen
        MCDA method name `mcda_name` for different modifications of criterion `j` weight.

        Parameters
        -----------
            matrix : ndarray
                Decision matrix with alternatives performances data. This matrix includes
                values of alternatives performances in rows considering criteria in columns

            weights : ndarray
                Vector with criteria weights. All weights in this vector must sum to 1.

            types : ndarray
                Vector with criteria types. Types can be equal to 1 for profit criteria and -1
                for cost criteria.

            percentages : ndarray
                Vector with percentage values of given criteria weight modification in range from 0 to 1.

            method : class
                Initialized object of class of chosen MCDA method

            j : int
                Index of column in decision matrix `matrix` that indicates for which criterion
                the weight is modified.

            dir_list : list
                list with directions (signs of value) of criterion weight modification. 1 denotes increasing,
                and -1 denotes decreasing weight value. You can provide [-1, 1] for increasing and
                decreasing, [-1] for only decreasing, or [1] for only increasing chosen criterion weight.

        Returns
        --------
            data_sens : DataFrame
                dataframe with rankings calculated for subsequent modifications of criterion j weight

        Examples
        ----------
        >>> sensitivity_analysis = Sensitivity_analysis_weights_percentages()
        >>> df_sens = sensitivity_analysis(matrix, weights, types, percentages, method, j, [-1, 1])
        """
        list_alt_names = [r'$A_{' + str(i) + '}$' for i in range(1, matrix.shape[0] + 1)]
        return Sensitivity_analysis_weights_percentages._sensitivity_analysis_weights_percentages(self, matrix, weights, types, percentages, method, list_alt_names, j, dir_list)


    def _change_weights(self, j, weights, change_val):
        """
        Method for criteria weights modification in sensitivity analysis procedure.

        Parameters
        -----------
            j : int
                Index of column in decision matrix `matrix` that indicates for which criterion
                the weight is modified.

            weights : ndarray
                Vector of criteria weights

            change_val : float
                Percentage value of criterion weight modification in range from 0 to 1

        Returns
        --------
            weights_copy : ndarray
                Vector with criteria weights after modification their values for sensitivity analysis

        """
        
        weights_copy = copy.deepcopy(weights)
        # Calculate value of selected criterion j weight modification
        change = weights_copy[j] * change_val
        # Calculate new value of selected criterion j weights
        new_weight = weights_copy[j] + change
        # Calculate new values of other criteria weights considering modification of j weights
        weights_copy = weights_copy - (change / (len(weights) - 1))
        # Assign new weight to criterion j
        weights_copy[j] = new_weight
        return weights_copy


    @staticmethod
    def _sensitivity_analysis_weights_percentages(self, matrix, weights, types, percentages, method, list_alt_names, j, dir_list):
        # Create a dataframe for sensitivity analysis results (rankings)
        data_sens = pd.DataFrame()
        # Assisgn indexes (alternatives symbols) to dataframe `data_sens`
        data_sens['Ai'] = list_alt_names
        # Iterate by two directions of weight modification: -1 for weight decreasing
        # and 1 for weight increasing
        for dir in dir_list:
            # Sorting percentages in appropriate order for visualization results
            if dir == -1:
                direct_percentages = copy.deepcopy(percentages[::-1])
            else:
                direct_percentages = copy.deepcopy(percentages)
            # Iterate by values of weight change in vector `direct_percentages`
            for change_val in direct_percentages:
                # Change weights using method named `_change_weights` from class `Sensitivity_analysis_weights_percentages`
                weights_copy = self._change_weights(j, weights, dir * change_val)

                # Calculate alternatives ranking using selected MCDA method, `matrix`, vector of new weights
                # `weights_copy` and criteria types `types`
                if method.__class__.__name__ == 'TOPSIS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'CODAS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'VIKOR':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = False)

                elif method.__class__.__name__ == 'SPOTIS':
                    bounds_min = np.amin(matrix, axis = 0)
                    bounds_max = np.amax(matrix, axis = 0)
                    bounds = np.vstack((bounds_min, bounds_max))
                    pref = method(matrix, weights_copy, types, bounds)
                    rank = rank_preferences(pref, reverse = False)

                elif method.__class__.__name__ == 'EDAS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'MABAC':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'MULTIMOORA':
                    rank = method(matrix, weights_copy, types)

                elif method.__class__.__name__ == 'WASPAS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'ARAS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'COPRAS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'CRADIS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'MARCOS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'CRADIS':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'SAW':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'PROMETHEE_II':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                elif method.__class__.__name__ == 'PROSA_C':
                    pref = method(matrix, weights_copy, types)
                    rank = rank_preferences(pref, reverse = True)

                else:
                    print('There is no available sensitivity analysis for this method')
                    return
                
                # Assign calculated ranking to column with value of j criterion weight change in `data_sens` dataframe
                data_sens['{:.0f}'.format(dir * change_val * 100) + '%'] = rank

        # Drop column with `-0%` name for visualization result
        if '-0%' in list(data_sens.columns):
            data_sens = data_sens.drop(['-0%'], axis = 1)
        # Set index with alternatives symbols in dataframe with sensitivity analysis results (rankings)
        data_sens = data_sens.set_index('Ai')
        return data_sens